from pathlib import Path

import PIL.Image
import PIL.ImageDraw
import PIL.ImageOps

import arcade
from arcade import hitbox
from arcade.cache import (
    HitBoxCache,
    ImageDataCache,
    TextureCache,
)
from arcade.texture import ImageData, SpriteSheet
from arcade.types.rect import Rect

from .texture import Texture


class TextureCacheManager:
    """
    A simple manager wrapping texture, image data and hit box caches
    with convenient methods for loading textures and sprite sheets.

    Args:
        hit_box_cache:
            Optional hit box cache to use. If not specified, a new cache will be created.
        image_data_cache:
            Optional image data cache to use. If not specified, a new cache will be created.
        texture_cache:
            Optional texture cache to use. If not specified, a new cache will be created
    """

    def __init__(
        self,
        hit_box_cache: HitBoxCache | None = None,
        image_data_cache: ImageDataCache | None = None,
        texture_cache: TextureCache | None = None,
    ):
        self._sprite_sheets: dict[str, SpriteSheet] = {}
        self._hit_box_cache = hit_box_cache or HitBoxCache()
        self._image_data_cache = image_data_cache or ImageDataCache()
        self._texture_cache = texture_cache or TextureCache()

    @property
    def hit_box_cache(self) -> HitBoxCache:
        """Get or set the current hit box cache."""
        return self._hit_box_cache

    @hit_box_cache.setter
    def hit_box_cache(self, value: HitBoxCache):
        self._hit_box_cache = value

    @property
    def image_data_cache(self) -> ImageDataCache:
        """Cache for image data."""
        return self._image_data_cache

    @property
    def texture_cache(self) -> TextureCache:
        """Cache for textures."""
        return self._texture_cache

    def flush(
        self,
        sprite_sheets: bool = True,
        textures: bool = True,
        image_data: bool = True,
        hit_boxes: bool = False,
    ):
        """
        Remove contents from the texture manager.

        Args:
            sprite_sheets:
                If ``True``, sprite sheets will be flushed.
            textures:
                If ``True``, textures will be flushed.
            image_data:
                If ``True``, image data will be flushed.
            hit_boxes:
                If ``True``, hit boxes will be flushed.
        """
        if sprite_sheets:
            self._sprite_sheets.clear()
        if textures:
            self._texture_cache.flush()
        if image_data:
            self._image_data_cache.flush()
        if hit_boxes:
            self._hit_box_cache.flush()

    def _get_real_path(self, path: str | Path) -> Path:
        """
        Resolve the path to the file.

        Args:
            path: Path to the file
        """
        if isinstance(path, str):
            return arcade.resources.resolve(path)
        elif isinstance(path, Path):
            return path
        else:
            raise TypeError(f"Invalid path type: {type(path)} for {path}")

    def load_or_get_spritesheet(self, path: str | Path) -> SpriteSheet:
        """
        Load a sprite sheet from disk, or return a cached version.

        Note that any texture sliced from the sprite sheet will be cached.
        if this is not the desirable behavior, use :meth:`load_or_get_spritesheet_texture`.

        Args:
            path: Path to the sprite sheet image
        """
        real_path = self._get_real_path(path)
        real_path_str = str(real_path)

        sprite_sheet = self._sprite_sheets.get(real_path_str)
        if sprite_sheet is None:
            sprite_sheet = arcade.SpriteSheet(real_path)
            self._sprite_sheets[real_path_str] = sprite_sheet

        return sprite_sheet

    def load_or_get_spritesheet_texture(
        self,
        path: str | Path,
        rect: Rect,
        hit_box_algorithm: hitbox.HitBoxAlgorithm | None = None,
    ) -> Texture:
        """
        Slice out a texture slice from a sprite sheet.

        * If the spritesheet is not already loaded, it will be loaded and cached.
        * If the sliced texture is already cached, it will be returned instead.

        Args:
            path:
                Path to the sprite sheet image
            rect:
                Slice of the texture in the sprite sheet.
            hit_box_algorithm:
                Hit box algorithm to use. If not specified, the global default will be used.
        """
        real_path = self._get_real_path(path)
        texture = self._texture_cache.get_texture_by_filepath(real_path, crop=rect.lbwh_int)
        if texture:
            return texture

        # check if sprite sheet is cached and load if not
        sprite_sheet = self.load_or_get_spritesheet(real_path)

        # slice out the texture and cache + return
        texture = sprite_sheet.get_texture(rect, hit_box_algorithm=hit_box_algorithm)
        self._texture_cache.put(texture)
        if texture.image_cache_name:
            self._image_data_cache.put(texture.image_cache_name, texture.image_data)

        # Add to image data cache
        self._image_data_cache.put(
            Texture.create_image_cache_name(real_path, rect.lbwh_int),
            texture.image_data,
        )

        return texture

    def load_or_get_image(
        self,
        path: str | Path,
        hash: str | None = None,
        mode="RGBA",
    ) -> ImageData:
        """
        Loads a complete image from disk or return a cached version.

        Args:
            path:
                Path of the file to load.
            hash:
                Optional override for image hash
            mode:
                The mode to use for the image. Default is "RGBA".
        """
        real_path = self._get_real_path(path)
        name = Texture.create_image_cache_name(real_path)
        im_data = self._image_data_cache.get(name)
        if im_data:
            return im_data
        image = PIL.Image.open(real_path).convert(mode)
        im_data = ImageData(image, hash=hash)
        self._image_data_cache.put(name, im_data)
        return im_data

    def load_or_get_texture(
        self,
        file_path: str | Path,
        *,
        x: int = 0,
        y: int = 0,
        width: int = 0,
        height: int = 0,
        hit_box_algorithm: hitbox.HitBoxAlgorithm | None = None,
    ) -> Texture:
        """
        Load an image from disk and create a texture. If the image is already
        loaded, return the cached version.

        The ``x``, ``y``, ``width``, and ``height`` parameters are used to
        specify a sub-rectangle of the image to load. If not specified, the
        entire image is loaded.

        Args:
            file_path:
                Path to the image file.
            x:
                X coordinate of the texture in the image.
            y:
                Y coordinate of the texture in the image.
            width:
                Width of the texture in the image.
            height:
                Height of the texture in the image.
            hit_box_algorithm:
                The hit box algorithm to use for this texture. If not specified,
                the global default will be used.
        """
        real_path = self._get_real_path(file_path)

        return self._load_or_get_texture(
            real_path,
            hit_box_algorithm=hit_box_algorithm,
            crop=(x, y, width, height),
        )

    def _load_or_get_texture(
        self,
        file_path: Path,
        hit_box_algorithm: hitbox.HitBoxAlgorithm | None = None,
        crop: tuple[int, int, int, int] = (0, 0, 0, 0),
        hash: str | None = None,
    ) -> Texture:
        """Load a texture, or return a cached version if it's already loaded."""
        hit_box_algorithm = hit_box_algorithm or hitbox.algo_default
        image_data: ImageData | None = None
        texture = None

        # Load the image data from disk or get from cache
        image_data, cached = self._load_or_get_image(file_path, hash=hash)
        # If the image was fetched from cache we might have cached texture
        if cached:
            texture = self._texture_cache.get_with_config(image_data.hash, hit_box_algorithm)
        # If we still don't have a texture, create it
        if texture is None:
            texture = Texture(image_data, hit_box_algorithm=hit_box_algorithm)
            texture.file_path = file_path
            texture.crop_values = crop
            self._texture_cache.put(texture)

        # If we have crop values we need to dig deeper looking for cached versions
        if crop != (0, 0, 0, 0):
            image_data = self._image_data_cache.get(
                Texture.create_image_cache_name(file_path, crop)
            )
            # If we don't have and cached image data we can crop from the base texture
            if image_data is None:
                texture = texture.crop(*crop)
                self._texture_cache.put(texture)
                self._image_data_cache.put(
                    Texture.create_image_cache_name(file_path, crop), texture.image_data
                )
            else:
                # We might have a texture for this image data
                texture = self._texture_cache.get_with_config(image_data.hash, hit_box_algorithm)
                if texture is None:
                    texture = Texture(image_data, hit_box_algorithm=hit_box_algorithm)
                    texture.file_path = file_path
                    texture.crop_values = crop
                    self._texture_cache.put(texture)

        return texture

    def _load_or_get_image(
        self,
        file_path: Path,
        hash: str | None = None,
        mode: str = "RGBA",
    ) -> tuple[ImageData, bool]:
        """
        Load an image, or return a cached version

        Args:
            file_path:
                Path to image
            hash:
                Hash of the image
            mode:
                The image mode to use (RGBA, RGB, etc.)

        Returns:
            Tuple of image data and a boolean indicating if the
            image was fetched from cache
        """
        file_path_str = str(file_path)
        cached = True

        # Do we have cached image data for this file?
        image_data = self._image_data_cache.get(Texture.create_image_cache_name(file_path_str))
        if not image_data:
            cached = False
            im = PIL.Image.open(file_path).convert(mode)
            image_data = ImageData(im, hash)
            self._image_data_cache.put(
                Texture.create_image_cache_name(file_path_str),
                image_data,
            )

        return image_data, cached



"""Constructs, are prepared widget combinations, you can use for common use-cases"""

from typing import Any

import arcade
from arcade import uicolor
from arcade.gui.events import UIOnActionEvent, UIOnClickEvent
from arcade.gui.mixins import UIMouseFilterMixin
from arcade.gui.nine_patch import NinePatchTexture
from arcade.gui.widgets.buttons import UIFlatButton
from arcade.gui.widgets.layout import UIAnchorLayout, UIBoxLayout
from arcade.gui.widgets.text import UILabel, UITextArea


class UIMessageBox(UIMouseFilterMixin, UIAnchorLayout):
    """A simple dialog box that pops up a message with buttons to close.
    Subclass this class or overwrite the 'on_action' event handler with

    .. code-block:: python

        box = UIMessageBox(...)
        @box.event("on_action")
        def on_action(event: UIOnActionEvent):
            pass

    Args:
      width: Width of the message box
      height: Height of the message box
      message_text: Text to show as message to the user
      title: Title of the message box, displayed on the top
      buttons: List of strings, which are shown as buttons

    """

    def __init__(
        self,
        *,
        width: float,
        height: float,
        message_text: str,
        title: str | None = None,
        buttons=("Ok",),
    ):
        if not buttons:
            raise ValueError("At least a single value has to be available for `buttons`")

        super().__init__(size_hint=(1, 1))
        self.register_event_type("on_action")
        self.with_background(color=uicolor.GRAY_CONCRETE.replace(a=150))

        space = 20

        # setup frame which will act like the window
        frame = self.add(UIAnchorLayout(width=width, height=height, size_hint=None))

        frame.with_background(
            texture=NinePatchTexture(
                left=7,
                right=7,
                bottom=7,
                top=7,
                texture=arcade.load_texture(":resources:gui_basic_assets/window/panel_gray.png"),
            )
        )

        # setup title
        if title:
            title_label = frame.add(
                child=UILabel(
                    text=title,
                    font_size=16,
                    size_hint=(1, 0),
                    align="center",
                ),
                anchor_y="top",
            )
            title_label.with_padding(all=2, bottom=5)
            title_label.with_background(color=uicolor.DARK_BLUE_MIDNIGHT_BLUE)
            title_offset = title_label.height
        else:
            title_offset = 0

        # Setup text
        text_area = frame.add(
            child=UITextArea(
                text=message_text,
                width=width - space,
                height=height - space,
                text_color=arcade.color.BLACK,
            ),
            anchor_x="center",
            anchor_y="top",
            align_y=-(title_offset + space),
        )
        text_area.with_padding(all=10)

        # setup buttons
        button_group = UIBoxLayout(vertical=False, space_between=10)
        for button_text in buttons:
            button = UIFlatButton(text=button_text)
            button_group.add(button)
            button.on_click = self._on_choice  # type: ignore

        frame.add(
            child=button_group, anchor_x="right", anchor_y="bottom", align_x=-space, align_y=space
        )

    def _on_choice(self, event):
        if self.parent:
            self.parent.remove(self)
        self.dispatch_event("on_action", UIOnActionEvent(self, event.source.text))

    def on_action(self, event: UIOnActionEvent):
        """Called when button was pressed"""
        pass


class UIButtonRow(UIBoxLayout):
    """Places buttons in a row.

    Args:
        vertical: Whether the button row is vertical or not.
        align: Where to align the button row.
        size_hint: Tuple of floats (0.0 - 1.0) of how much space of the
            parent should be requested.
        size_hint_min: Min width and height in pixel.
        size_hint_max: Max width and height in pixel.
        space_between: The space between the children.
        callback: The callback function which will receive the text of
            the clicked button.
        button_factory: The factory to create the buttons. Default is py:class:`UIFlatButton`.
        **kwargs: Passed to UIBoxLayout
    """

    def __init__(
        self,
        *,
        vertical: bool = False,
        align: str = "center",
        size_hint: Any = (0, 0),
        size_hint_min: Any | None = None,
        size_hint_max: Any | None = None,
        space_between: int = 10,
        button_factory: type = UIFlatButton,
        **kwargs,
    ):
        super().__init__(
            vertical=vertical,
            align=align,
            size_hint=size_hint,
            size_hint_min=size_hint_min,
            size_hint_max=size_hint_max,
            space_between=space_between,
            **kwargs,
        )
        self.register_event_type("on_action")

        self.button_factory = button_factory

    def add_button(self, label: str, *, style=None, multiline=False, **kwargs):
        """Add a button to the row.

        Args:
            label: The text of the button.
            style: The style of the button.
            multiline: Whether the button is multiline or not.
            **kwargs: Passed to the button factory.
        """
        button = self.button_factory(text=label, style=style, multiline=multiline, **kwargs)
        button.on_click = self._on_click
        self.add(button)
        return button

    def on_action(self, event: UIOnActionEvent):
        """Called when button was pressed, override this method to handle button presses."""
        pass

    def _on_click(self, event: UIOnClickEvent):
        self.dispatch_event("on_action", UIOnActionEvent(self, event.source.text))



"""
Drawing bezier curve using tessellation shader.
This example is ported from moderngl : https://github.com/moderngl/moderngl/blob/master/examples/tesselation.py

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.tessellation
"""

from array import array

import arcade
from arcade.gl import BufferDescription

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Bezier Curve Tessellation"


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        """
        Set up the application.
        """
        super().__init__(width, height, title, gl_version=(4, 1))
        self.program = self.ctx.program(
            vertex_shader="""
            #version 400 core

            uniform float time;
            in vec2 in_vert;

            void main() {
                vec2 pos = vec2(in_vert.x, sin(in_vert.y + in_vert.x * 1.5 + time * 0.5));
                gl_Position = vec4(pos, 0.0, 1.0);
            }
            """,
            tess_control_shader="""
            #version 400 core

            uniform float time;
            layout(vertices = 4) out;

            void main() {
                // set tesselation levels (can be computed dynamically)
                gl_TessLevelOuter[0] = 1;
                gl_TessLevelOuter[1] = int(time * 2) % 32 + 1;
                // pass through vertex positions
                gl_out[gl_InvocationID].gl_Position = gl_in[gl_InvocationID].gl_Position;
            }
            """,
            tess_evaluation_shader="""
            #version 400 core

            layout(isolines, fractional_even_spacing, ccw) in;

            // compute a point on a bezier curve with the points p0, p1, p2, p3
            // the parameter u is in [0, 1] and determines the position on the curve
            vec3 bezier(float u, vec3 p0, vec3 p1, vec3 p2, vec3 p3) {
                float B0 = (1.0 - u) * (1.0 - u) * (1.0 - u);
                float B1 = 3.0 * (1.0 - u) * (1.0 - u) * u;
                float B2 = 3.0 * (1.0 - u) * u * u;
                float B3 = u * u * u;
                return B0 * p0 + B1 * p1 + B2 * p2 + B3 * p3;
            }
            void main() {
                float u = gl_TessCoord.x;
                vec3 p0 = vec3(gl_in[0].gl_Position);
                vec3 p1 = vec3(gl_in[1].gl_Position);
                vec3 p2 = vec3(gl_in[2].gl_Position);
                vec3 p3 = vec3(gl_in[3].gl_Position);
                gl_Position = vec4(bezier(u, p0, p1, p2, p3), 1.0);
            }
            """,
            fragment_shader="""
            #version 400 core

            out vec4 frag_color;

            void main() {
                frag_color = vec4(1.0);
            }
            """,
        )

        # We are processing patches of 4 vertices
        self.ctx.patch_vertices = 4

        # fmt: off
        vertices = array(
            "f",
            # x, y
            [
                -1.0, 0.0,
                -0.5, 1.0,
                0.5, -1.0,
                1.0, 0.0,
            ]
        )
        # fmt: on

        self.geometry = self.ctx.geometry(
            [BufferDescription(self.ctx.buffer(data=vertices), "2f", ["in_vert"])]
        )

    def on_draw(self):
        self.clear()
        self.program["time"] = self.time
        self.geometry.render(self.program, mode=self.ctx.PATCHES)


if __name__ == "__main__":
    GameView(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    arcade.run()



"""
Scroll around a large screen.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_move_scrolling
"""

import random
import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Move with Scrolling Screen Example"

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = 220

# How fast the camera pans to the player. 1.0 is instant.
CAMERA_SPEED = 0.1

# How fast the character moves
PLAYER_MOVEMENT_SPEED = 7


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Sprite lists
        self.player_list = None
        self.wall_list = None

        # Set up the player
        self.player_sprite = None

        # Physics engine so we don't run into walls.
        self.physics_engine = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

        # Create the cameras. One for the GUI, one for the sprites.
        # We scroll the 'sprite world' but not the GUI.
        self.camera_sprites = arcade.Camera2D()
        self.camera_gui = arcade.Camera2D()

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.4,
        )
        self.player_sprite.center_x = 256
        self.player_sprite.center_y = 512
        self.player_list.append(self.player_sprite)

        # -- Set up several columns of walls
        for x in range(200, 1650, 210):
            for y in range(0, 1600, 64):
                # Randomly skip a box so the player can find a way through
                if random.randrange(5) > 0:
                    wall = arcade.Sprite(
                        ":resources:images/tiles/grassCenter.png",
                        scale=SPRITE_SCALING,
                    )
                    wall.center_x = x
                    wall.center_y = y
                    self.wall_list.append(wall)

        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite, self.wall_list)

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self):
        """ Render the screen. """

        # This command has to happen before we start drawing
        self.clear()

        # Select the camera we'll use to draw all our sprites
        self.camera_sprites.use()

        # Draw all the sprites.
        self.wall_list.draw()
        self.player_list.draw()

        # Select the (unscrolled) camera for our GUI
        self.camera_gui.use()

        # Draw the GUI
        arcade.draw_rect_filled(arcade.rect.XYWH(self.width // 2, 20, self.width, 40),
                                arcade.color.ALMOND)
        text = f"Scroll value: ({self.camera_sprites.position[0]:5.1f}, " \
               f"{self.camera_sprites.position[1]:5.1f})"
        arcade.draw_text(text, 10, 10, arcade.color.BLACK_BEAN, 20)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.up_pressed = True
        elif key == arcade.key.DOWN:
            self.down_pressed = True
        elif key == arcade.key.LEFT:
            self.left_pressed = True
        elif key == arcade.key.RIGHT:
            self.right_pressed = True

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP:
            self.up_pressed = False
        elif key == arcade.key.DOWN:
            self.down_pressed = False
        elif key == arcade.key.LEFT:
            self.left_pressed = False
        elif key == arcade.key.RIGHT:
            self.right_pressed = False

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Calculate speed based on the keys pressed
        self.player_sprite.change_x = 0
        self.player_sprite.change_y = 0

        if self.up_pressed and not self.down_pressed:
            self.player_sprite.change_y = PLAYER_MOVEMENT_SPEED
        elif self.down_pressed and not self.up_pressed:
            self.player_sprite.change_y = -PLAYER_MOVEMENT_SPEED
        if self.left_pressed and not self.right_pressed:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        elif self.right_pressed and not self.left_pressed:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()

        # Scroll the screen to the player
        self.scroll_to_player()

    def scroll_to_player(self):
        """
        Scroll the window to the player.

        if CAMERA_SPEED is 1, the camera will immediately move to the desired
        position. Anything between 0 and 1 will have the camera move to the
        location with a smoother pan.
        """

        position = (self.player_sprite.center_x, self.player_sprite.center_y)
        self.camera_sprites.position = arcade.math.lerp_2d(
            self.camera_sprites.position, position, CAMERA_SPEED,
        )

    def on_resize(self, width: int, height: int):
        """
        Resize window
        Handle the user grabbing the edge and resizing the window.
        """
        super().on_resize(width, height)
        self.camera_sprites.match_window()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example demonstrating a grid layout with focusable buttons in an Arcade GUI.

This example shows how to create a grid layout with buttons
that can be navigated using a controller.
It includes a focus transition setup to allow smooth navigation between buttons in the grid.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_controller_support_grid
"""


import arcade
from arcade.examples.gui.exp_controller_support import ControllerIndicator
from arcade.experimental.controller_window import ControllerView, ControllerWindow
from arcade.gui import (
    UIFlatButton,
    UIGridLayout,
    UIView,
    UIWidget,
)
from arcade.gui.experimental.focus import UIFocusable, UIFocusGroup


class FocusableButton(UIFocusable, UIFlatButton):
    pass


def setup_grid_focus_transition(grid: dict[tuple[int, int], UIWidget]):
    """Setup focus transition in grid.

    Connect focus transition between `Focusable` in grid.

    Args:
        grid: Dict[Tuple[int, int], Focusable]: grid of Focusable widgets.
        key represents position in grid (x,y)

    """

    cols = max(x for x, y in grid.keys()) + 1
    rows = max(y for x, y in grid.keys()) + 1
    for c in range(cols):
        for r in range(rows):
            btn = grid.get((c, r))
            if btn is None or not isinstance(btn, UIFocusable):
                continue

            if c > 0:
                btn.neighbor_left = grid.get((c - 1, r))
            else:
                btn.neighbor_left = grid.get((cols - 1, r))

            if c < cols - 1:
                btn.neighbor_right = grid.get((c + 1, r))
            else:
                btn.neighbor_right = grid.get((0, r))

            if r > 0:
                btn.neighbor_up = grid.get((c, r - 1))
            else:
                btn.neighbor_up = grid.get((c, rows - 1))

            if r < rows - 1:
                btn.neighbor_down = grid.get((c, r + 1))
            else:
                btn.neighbor_down = grid.get((c, 0))


class MyView(ControllerView, UIView):
    def __init__(self):
        super().__init__()
        arcade.set_background_color(arcade.color.AMAZON)

        self.root = self.add_widget(ControllerIndicator())
        self.root = self.root.add(UIFocusGroup())
        grid = self.root.add(
            UIGridLayout(column_count=3, row_count=3, vertical_spacing=10, horizontal_spacing=10)
        )

        _grid = {}
        for i in range(9):
            btn = FocusableButton(text=f"Button {i}")
            _grid[(i % 3, i // 3)] = btn
            grid.add(btn, column=i % 3, row=i // 3)

        # connect focus transition in grid
        setup_grid_focus_transition(_grid)

        self.root.detect_focusable_widgets()


if __name__ == "__main__":
    window = ControllerWindow(title="Controller UI Example")
    window.show_view(MyView())
    arcade.run()



"""
Acceleration and Friction

Demonstrate how to implement simple acceleration and friction without a
physics engine.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.sprite_move_keyboard_accel
"""

import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Better Move Sprite with Keyboard Example"

# Important constants for this example

# Speed limit
MAX_SPEED = 3.0

# How fast we accelerate
ACCELERATION_RATE = 0.1

# How fast to slow down after we let off the key
FRICTION = 0.02


class Player(arcade.Sprite):

    def update(self, delta_time: float = 1/60):
        self.center_x += self.change_x
        self.center_y += self.change_y

        # Check to see if we hit the screen edge
        if self.left < 0:
            self.left = 0
            self.change_x = 0  # Zero x speed
        elif self.right > WINDOW_WIDTH - 1:
            self.right = WINDOW_WIDTH - 1
            self.change_x = 0

        if self.bottom < 0:
            self.bottom = 0
            self.change_y = 0
        elif self.top > WINDOW_HEIGHT - 1:
            self.top = WINDOW_HEIGHT - 1
            self.change_y = 0


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variable to will hold the player sprite list
        self.player_list = None

        # Create a place to store the player sprite
        # so it can be  accessed directly.
        self.player_sprite = None

        # Create places to store the speed display Text objects
        self.x_speed_display = None
        self.y_speed_display = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Create a sprite list
        self.player_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = Player(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=SPRITE_SCALING,
        )
        self.player_sprite.position = self.width / 2, self.height / 2
        self.player_list.append(self.player_sprite)

        # Create the speed display objects with initial text
        self.x_speed_display = arcade.Text(
            f"X Speed: {self.player_sprite.change_x:6.3f}",
            10, 50, arcade.color.BLACK, font_size=15,
        )

        self.y_speed_display = arcade.Text(
            f"Y Speed: {self.player_sprite.change_y:6.3f}",
            10, 70, color=arcade.color.BLACK, font_size=15,
        )

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites.
        self.player_list.draw()

        # Draw the speed indicators
        self.x_speed_display.draw()
        self.y_speed_display.draw()

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Add some friction
        if self.player_sprite.change_x > FRICTION:
            self.player_sprite.change_x -= FRICTION
        elif self.player_sprite.change_x < -FRICTION:
            self.player_sprite.change_x += FRICTION
        else:
            self.player_sprite.change_x = 0

        if self.player_sprite.change_y > FRICTION:
            self.player_sprite.change_y -= FRICTION
        elif self.player_sprite.change_y < -FRICTION:
            self.player_sprite.change_y += FRICTION
        else:
            self.player_sprite.change_y = 0

        # Apply acceleration based on the keys pressed
        if self.up_pressed and not self.down_pressed:
            self.player_sprite.change_y += ACCELERATION_RATE
        elif self.down_pressed and not self.up_pressed:
            self.player_sprite.change_y += -ACCELERATION_RATE
        if self.left_pressed and not self.right_pressed:
            self.player_sprite.change_x += -ACCELERATION_RATE
        elif self.right_pressed and not self.left_pressed:
            self.player_sprite.change_x += ACCELERATION_RATE

        if self.player_sprite.change_x > MAX_SPEED:
            self.player_sprite.change_x = MAX_SPEED
        elif self.player_sprite.change_x < -MAX_SPEED:
            self.player_sprite.change_x = -MAX_SPEED
        if self.player_sprite.change_y > MAX_SPEED:
            self.player_sprite.change_y = MAX_SPEED
        elif self.player_sprite.change_y < -MAX_SPEED:
            self.player_sprite.change_y = -MAX_SPEED

        # Call update to move the sprite
        # IMPORTANT: If using a physics engine, you need to call update
        # on it instead of the sprite list!
        self.player_list.update(delta_time)

        # Update the speed displays based on the final speed
        self.x_speed_display.text = f"X Speed: {self.player_sprite.change_x:6.3f}"
        self.y_speed_display.text = f"Y Speed: {self.player_sprite.change_y:6.3f}"

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.up_pressed = True
        elif key == arcade.key.DOWN:
            self.down_pressed = True
        elif key == arcade.key.LEFT:
            self.left_pressed = True
        elif key == arcade.key.RIGHT:
            self.right_pressed = True

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP:
            self.up_pressed = False
        elif key == arcade.key.DOWN:
            self.down_pressed = False
        elif key == arcade.key.LEFT:
            self.left_pressed = False
        elif key == arcade.key.RIGHT:
            self.right_pressed = False


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""Example of creating a custom progress bar.

This example shows how to create a custom layout.

The layout will place the widgets in a rotating circle around the center of the screen.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.own_layout
"""

from math import cos, sin
from typing import TypeVar

import arcade
from arcade.gui import UIAnchorLayout, UIFlatButton, UILayout, UIView, UIWidget

W = TypeVar("W", bound="UIWidget")


class CircleLayout(UILayout):
    """A custom progress bar widget.

    A UIAnchorLayout is a layout that arranges its children in a specific way.
    The actual bar is a UISpace that fills the parent widget from left to right.
    """

    def __init__(self, size_hint=(1, 1), **kwargs):
        super().__init__(size_hint=size_hint, **kwargs)
        self._time = 0  # used for rotation

    def add(self, child: W, **kwargs) -> W:
        """Add a widget to the layout.

        The widget is placed in a circle around the center of the screen.
        """
        return super().add(child, **kwargs)

    def do_layout(self):
        """Layout the children in a circle around the center of the screen."""
        if not self._children:
            return

        # handle the size hints of the children
        for child in self.children:
            self._resize_child(child)

        # calculate the radius, so that the children are placed inside the parent widget
        max_child_width = max(child.content_width for child in self.children)
        max_child_height = max(child.content_height for child in self.children)
        radius = (
            min(self.content_width - max_child_width, self.content_height - max_child_height) / 2
        )

        for i, child in enumerate(self.children):
            angle = i / len(self.children) * 2 * 3.1415
            # add rotation based on time
            angle += self._time * 0.08
            center_x = self.center_x + radius * cos(angle)
            center_y = self.center_y + radius * sin(angle)

            new_rect = child.rect.align_center((center_x, center_y))
            child.rect = new_rect

    def _resize_child(self, child: UIWidget):
        """Resizes the child based on the size_hint, size_hint_min, and size_hint_max."""
        new_child_rect = child.rect

        sh_w, sh_h = child.size_hint or (None, None)
        shmn_w, shmn_h = child.size_hint_min or (None, None)
        shmx_w, shmx_h = child.size_hint_max or (None, None)

        if sh_w is not None:
            new_child_rect = new_child_rect.resize(width=self.content_width * sh_w)

            if shmn_w:
                new_child_rect = new_child_rect.min_size(width=shmn_w)
            if shmx_w:
                new_child_rect = new_child_rect.max_size(width=shmx_w)

        if sh_h is not None:
            new_child_rect = new_child_rect.resize(height=self.content_height * sh_h)

            if shmn_h:
                new_child_rect = new_child_rect.min_size(height=shmn_h)
            if shmx_h:
                new_child_rect = new_child_rect.max_size(height=shmx_h)

        child.rect = new_child_rect

    def on_update(self, dt):
        self._time += dt


class MyView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        root = self.ui.add(UIAnchorLayout())

        # Create a custom layout
        self.circle_layout = root.add(CircleLayout(size_hint=(0.8, 0.8)))

        # Add buttons to the layout
        for i in range(8):
            self.circle_layout.add(
                UIFlatButton(
                    text=f"Button {i}",
                    size_hint=(0.1, 0.1),
                )
            )


def main():
    window = arcade.Window(title="GUI Example: CircleLayout")
    window.show_view(MyView())
    arcade.run()


if __name__ == "__main__":
    main()



"""
Interpolate a sprite's motion which is calculated in fixed update.

The bouncing done in this example is very bare-bones, and unstable.
The tick speed of the global clock has been slowed down.
This helps highlight the natural 'laggy' behaviour of fixed update physics
the fixed update should be kept close to the nominal update rate, or even faster.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.fixed_update_interpolation.py
"""
import arcade
import arcade.clock

# --- Constants ---
GRAVITY = 98.1  # 98.1 px per second
CIRCLE_RADIUS = 30

WINDOW_WIDTH = 1289
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Follow Path Simple Example"


class GameView(arcade.View):

    def __init__(self):
        super().__init__()
        self.unfixed_sprite = arcade.SpriteCircle(CIRCLE_RADIUS, arcade.color.RADICAL_RED)
        self.interpolated_sprite = arcade.SpriteCircle(CIRCLE_RADIUS, arcade.color.ORANGE)
        self.fixed_sprite = arcade.SpriteCircle(CIRCLE_RADIUS, arcade.color.GOLD)

        # by setting the tick speed to 0.1 the standard update
        # perseves time at a tenth speed
        arcade.clock.GLOBAL_CLOCK.set_tick_speed(0.1)

        # We store the last position of the fixed sprite to find
        # the interpolated sprite's position
        self.last_position = 0.0

        self.sprites = arcade.SpriteList()
        self.sprites.extend((self.unfixed_sprite, self.fixed_sprite, self.interpolated_sprite))

    def setup(self):
        self.unfixed_sprite.change_y = self.interpolated_sprite.change_y = 0.0
        self.fixed_sprite.change_y = self.interpolated_sprite.change_y = 0.0

        self.unfixed_sprite.position = WINDOW_WIDTH / 4.0, WINDOW_HEIGHT / 2.0
        self.interpolated_sprite.position = 2.0 * WINDOW_WIDTH / 4.0, WINDOW_HEIGHT / 2.0
        self.fixed_sprite.position = 3.0 * WINDOW_WIDTH / 4.0, WINDOW_HEIGHT / 2.0

        self.last_position = self.fixed_sprite.center_y

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == arcade.key.R:
            self.setup()

    def on_fixed_update(self, delta_time: float):
        # Accelerate the sprite downward due to gravity
        self.fixed_sprite.change_y -= GRAVITY * delta_time

        # If the sprite is colliding with the ground then make it 'bounce' by
        # flipping it's velocity
        if (
            self.fixed_sprite.center_y <= CIRCLE_RADIUS
            and self.fixed_sprite.change_y <= 0.0
        ):
            self.fixed_sprite.change_y *= -1

        # Move the sprite based on its velocity
        self.last_position = self.fixed_sprite.center_y
        self.fixed_sprite.center_y += self.fixed_sprite.change_y * delta_time

    def on_update(self, delta_time: float):
        # Accelerate the sprite downward due to gravity
        self.unfixed_sprite.change_y -= GRAVITY * delta_time

        # If the sprite is colliding with the ground then make it 'bounce'
        # by flipping it's velocity
        if (
            self.unfixed_sprite.center_y <= CIRCLE_RADIUS
            and self.unfixed_sprite.change_y <= 0.0
        ):
            self.unfixed_sprite.change_y *= -1

        # Move the sprite based on its velocity
        self.unfixed_sprite.center_y += self.unfixed_sprite.change_y * delta_time

        self.interpolated_sprite.center_y = arcade.math.lerp(
            self.last_position,
            self.fixed_sprite.center_y,
            arcade.clock.GLOBAL_FIXED_CLOCK.fraction,
        )

    def on_draw(self):
        self.clear()
        self.sprites.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, fixed_rate=1/60)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == '__main__':
    main()



"""
Simple Snow
Based primarily on:
https://api.arcade.academy/en/latest/examples/sprite_collect_coins_move_down.html

Contributed to Python Arcade Library by Nicholas Hartunian

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.snow
"""

import random
import math
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Snow"
SNOWFLAKE_COUNT = 800


class Snowflake(arcade.SpriteCircle):
    """
    Each instance of this class represents a single snowflake.
    Based on drawing filled-circles.
    """

    def __init__(self, size, speed, drift):
        super().__init__(size, arcade.color.WHITE)
        self.speed = speed
        self.drift = drift

    def reset_pos(self):
        # Reset flake to random position above screen
        self.position = (
            random.randrange(WINDOW_WIDTH),
            random.randrange(WINDOW_HEIGHT, WINDOW_HEIGHT + 100),
        )

    def update(self, delta_time: float = 1/60) -> None:
        self.center_y -= self.speed * delta_time

        # Check if snowflake has fallen below screen
        if self.center_y < 0:
            self.reset_pos()

        # Some math to make the snowflakes move side to side
        self.center_x += self.speed * math.cos(self.drift) * delta_time
        self.drift += 1 * delta_time


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """ Initializer """
        # Calls "__init__" of parent class (arcade.Window) to setup screen
        super().__init__()

        self.snowflake_list = arcade.SpriteList()

        # Don't show the mouse pointer
        self.window.set_mouse_visible(False)

        # Set the background color
        self.background_color = arcade.color.BLACK

    def start_snowfall(self):
        """ Set up snowfall and initialize variables. """
        for i in range(SNOWFLAKE_COUNT):
            # Create snowflake instance
            snowflake = Snowflake(
                size=random.randrange(1, 4),
                speed=random.randrange(20, 40),
                drift=random.uniform(math.pi, math.pi * 2),
            )
            # Randomly position snowflake
            snowflake.position = (
                random.randrange(WINDOW_WIDTH),
                random.randrange(WINDOW_HEIGHT + 200),
            )
            # Add snowflake to snowflake list
            self.snowflake_list.append(snowflake)

    def on_draw(self):
        """ Render the screen. """
        # Clear the screen to the background color
        self.clear()

        # Draw the current position of each snowflake
        self.snowflake_list.draw()

    def on_update(self, delta_time):
        """ All the logic to move, and the game logic goes here. """
        # Call update on all the snowflakes
        self.snowflake_list.update(delta_time)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.start_snowfall()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
A module providing commonly used geometry
"""

import math
from array import array

from arcade.gl import BufferDescription, Context
from arcade.gl.vertex_array import Geometry


def _get_active_context() -> Context:
    """
    Get the global active context.

    The gl module is forbidden to reach outside of its own module.
    so we can't use ``arcade.get_window().ctx`` here.
    """
    ctx = Context.active
    if not ctx:
        raise RuntimeError("No context is currently activated")
    return ctx


def quad_2d_fs() -> Geometry:
    """Creates a screen aligned quad using normalized device coordinates"""
    return quad_2d(size=(2.0, 2.0))


def quad_2d(
    size: tuple[float, float] = (1.0, 1.0), pos: tuple[float, float] = (0.0, 0.0)
) -> Geometry:
    """
    Creates 2D quad Geometry using 2 triangle strip with texture coordinates.

    By default a unit quad is created with the center at (0, 0).

    Args:
        size:
            width and height of the quad
        pos:
            Center position x and y
    """
    ctx = _get_active_context()
    width, height = size
    x_pos, y_pos = pos

    # fmt: off
    data = array('f', [
        x_pos - width / 2.0, y_pos + height / 2.0, 0.0, 1.0,
        x_pos - width / 2.0, y_pos - height / 2.0, 0.0, 0.0,
        x_pos + width / 2.0, y_pos + height / 2.0, 1.0, 1.0,
        x_pos + width / 2.0, y_pos - height / 2.0, 1.0, 0.0,
    ])
    # fmt: on

    return ctx.geometry(
        [
            BufferDescription(
                ctx.buffer(data=data),
                "2f 2f",
                ["in_vert", "in_uv"],
            )
        ],
        mode=ctx.TRIANGLE_STRIP,
    )


def screen_rectangle(
    bottom_left_x: float, bottom_left_y: float, width: float, height: float
) -> Geometry:
    """
    Creates screen rectangle using 2 triangle strip with texture coordinates.

    This can create a rectangle in normalized device coordinates or projection space.

    Args:
        bottom_left_x:
            Bottom left x position
        bottom_left_y:
            Bottom left y position
        width:
            Width of the rectangle
        height:
            Height of the rectangle
    """
    ctx = _get_active_context()
    # fmt: off
    data = array('f', [
        bottom_left_x, bottom_left_y + height, 0.0, 1.0,
        bottom_left_x, bottom_left_y, 0.0, 0.0,
        bottom_left_x + width, bottom_left_y + height, 1.0, 1.0,
        bottom_left_x + width, bottom_left_y, 1.0, 0.0,
    ])
    # fmt: on
    return ctx.geometry(
        [
            BufferDescription(
                ctx.buffer(data=data),
                "2f 2f",
                ["in_vert", "in_uv"],
            )
        ],
        mode=ctx.TRIANGLE_STRIP,
    )


def cube(
    size: tuple[float, float, float] = (1.0, 1.0, 1.0),
    center: tuple[float, float, float] = (0.0, 0.0, 0.0),
) -> Geometry:
    """
    Creates a cube with normals and texture coordinates.

    By default a unit cube is created with the center at (0, 0, 0).

    Args:
        size:
            Size of the cube as a 3-component tuple
        center:
            Center of the cube as a 3-component tuple
    Returns:
        A Geometry instance containing a cube
    """
    ctx = _get_active_context()
    width, height, depth = size
    width, height, depth = width / 2.0, height / 2.0, depth / 2.0

    # fmt: off
    pos = array('f', [
        center[0] + width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] + height, center[2] + depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] + height, center[2] + depth,
        center[0] - width, center[1] + height, center[2] + depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] - height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] - depth,
        center[0] + width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] + height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] + depth,
        center[0] + width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] - height, center[2] - depth,
        center[0] + width, center[1] - height, center[2] + depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] + width, center[1] - height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] - width, center[1] - height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] - width, center[1] + height, center[2] + depth,
        center[0] - width, center[1] + height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] + depth,
        center[0] - width, center[1] + height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] - depth,
        center[0] + width, center[1] - height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] - depth,
        center[0] - width, center[1] - height, center[2] - depth,
        center[0] - width, center[1] + height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] - depth,
        center[0] - width, center[1] + height, center[2] - depth,
        center[0] + width, center[1] + height, center[2] + depth,
        center[0] - width, center[1] + height, center[2] - depth,
        center[0] - width, center[1] + height, center[2] + depth,
        center[0] + width, center[1] + height, center[2] + depth,
    ])
    normal = array('f', [
        -0, 0, 1,
        -0, 0, 1,
        -0, 0, 1,
        0, 0, 1,
        0, 0, 1,
        0, 0, 1,
        1, 0, 0,
        1, 0, 0,
        1, 0, 0,
        1, 0, 0,
        1, 0, 0,
        1, 0, 0,
        0, -1, 0,
        0, -1, 0,
        0, -1, 0,
        0, -1, 0,
        0, -1, 0,
        0, -1, 0,
        -1, -0, 0,
        -1, -0, 0,
        -1, -0, 0,
        -1, -0, 0,
        -1, -0, 0,
        -1, -0, 0,
        0, 0, -1,
        0, 0, -1,
        0, 0, -1,
        0, 0, -1,
        0, 0, -1,
        0, 0, -1,
        0, 1, 0,
        0, 1, 0,
        0, 1, 0,
        0, 1, 0,
        0, 1, 0,
        0, 1, 0,
    ])

    uv = array('f', [
        1, 0,
        1, 1,
        0, 0,
        1, 1,
        0, 1,
        0, 0,
        1, 0,
        1, 1,
        0, 0,
        1, 1,
        0, 1,
        0, 0,
        1, 1,
        0, 1,
        0, 0,
        1, 1,
        0, 0,
        1, 0,
        0, 1,
        0, 0,
        1, 0,
        0, 1,
        1, 0,
        1, 1,
        1, 0,
        1, 1,
        0, 1,
        1, 0,
        0, 1,
        0, 0,
        1, 1,
        0, 1,
        1, 0,
        0, 1,
        0, 0,
        1, 0
    ])
    # fmt: on

    return ctx.geometry(
        [
            BufferDescription(ctx.buffer(data=pos), "3f", ["in_position"]),
            BufferDescription(ctx.buffer(data=normal), "3f", ["in_normal"]),
            BufferDescription(ctx.buffer(data=uv), "2f", ["in_uv"]),
        ]
    )


def sphere(
    radius=0.5,
    sectors=32,
    rings=16,
    normals=True,
    uvs=True,
) -> Geometry:
    """
    Creates a 3D sphere.

    Args:
        radius:
            Radius or the sphere
        rings:
            number or horizontal rings
        sectors:
            number of vertical segments
        normals:
            Include normals
        uvs:
            Include texture coordinates
    Returns:
        A Geometry instance containing a sphere
    """
    ctx = _get_active_context()

    R = 1.0 / (rings - 1)
    S = 1.0 / (sectors - 1)

    vertices = [0.0] * (rings * sectors * 3)
    normals = [0.0] * (rings * sectors * 3)
    uvs = [0.0] * (rings * sectors * 2)

    v, n, t = 0, 0, 0
    for r in range(rings):
        for s in range(sectors):
            y = math.sin(-math.pi / 2 + math.pi * r * R)
            x = math.cos(2 * math.pi * s * S) * math.sin(math.pi * r * R)
            z = math.sin(2 * math.pi * s * S) * math.sin(math.pi * r * R)

            uvs[t] = s * S
            uvs[t + 1] = r * R

            vertices[v] = x * radius
            vertices[v + 1] = y * radius
            vertices[v + 2] = z * radius

            normals[n] = x
            normals[n + 1] = y
            normals[n + 2] = z

            t += 2
            v += 3
            n += 3

    indices = [0] * rings * sectors * 6
    i = 0
    for r in range(rings - 1):
        for s in range(sectors - 1):
            indices[i] = r * sectors + s
            indices[i + 1] = (r + 1) * sectors + (s + 1)
            indices[i + 2] = r * sectors + (s + 1)

            indices[i + 3] = r * sectors + s
            indices[i + 4] = (r + 1) * sectors + s
            indices[i + 5] = (r + 1) * sectors + (s + 1)
            i += 6

    content = [
        BufferDescription(ctx.buffer(data=array("f", vertices)), "3f", ["in_position"]),
    ]
    if normals:
        content.append(BufferDescription(ctx.buffer(data=array("f", normals)), "3f", ["in_normal"]))
    if uvs:
        content.append(BufferDescription(ctx.buffer(data=array("f", uvs)), "2f", ["in_uv"]))

    return ctx.geometry(
        content,
        index_buffer=ctx.buffer(data=array("I", indices)),
        mode=ctx.TRIANGLES,
    )



"""
A very simple example rendering a textured fullscreen rectangle.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.shader_setup_simple_texture
"""

import arcade
from arcade.gl import geometry

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Shader Setup"


class ShaderSetup(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            // The vertex shader will run for every vertex of the input geometry
            // and set the final position of a vertex.
            // In this example the rectangle geometry we are rendering
            // consists of two triangles.

            // The expected inputs names the geometry module are using
            in vec2 in_vert;
            in vec2 in_uv;

            // Texture coordinate output to the fragment shader
            out vec2 v_uv;

            void main() {
                // Set the vertex position. This is a special out variable for positions.s
                gl_Position = vec4(in_vert, 0.0, 1.0);
                // Send the texture coordinates to the fragment shader
                v_uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            // The fragment shader will run for every pixel it needs
            // two fill in the two triangles we are rendering.
            // The texture coordinate passed in will be an interpolated
            // value between the texture coordinate specified in each vertex.
            // It interpolates based on location of the current pixel in the triangle.

            // A sampler that can read a texture from a channel
            uniform sampler2D background;

            // Texture coordinate from the vertex shader
            in vec2 v_uv;

            // The final pixel value that will be written to the screen
            out vec4 out_color;

            void main() {
                // Read a pixel from current texture coordinate and write that to the screen.
                out_color = texture(background, v_uv);
            }
            """,
        )
        # Configure the sampler to read from texture channel 0.
        # Sampler uniforms are simply integers containing what
        # texture channel to read from. Most hardware have 8 to 16 channels.
        self.program["background"] = 0
        # Create geometry for a fullscreen rectangle in normalized device coordinates
        self.quad = geometry.screen_rectangle(-1, -1, 2, 2)
        # Load a texture we want to manipulate. This is an OpenGL texture, not an arcade.Texture
        self.texture = self.ctx.load_texture(":resources:images/backgrounds/abstract_1.jpg")

    def on_draw(self):
        self.clear()
        # Bind the texture channel 0 (we configured the sampler to use channel 0)
        self.texture.use(0)
        # Draw the geometry using the program
        self.quad.render(self.program)


ShaderSetup().run()



"""An overview of all included widgets.

See the other GUI examples for more indepth information about specific widgets.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.2_widgets
"""

import textwrap
from copy import deepcopy

import arcade
from arcade.gui import (
    NinePatchTexture,
    UIAnchorLayout,
    UIBoxLayout,
    UIButtonRow,
    UIDropdown,
    UIDummy,
    UIFlatButton,
    UIImage,
    UIInputText,
    UILabel,
    UIManager,
    UIMessageBox,
    UIOnActionEvent,
    UIOnChangeEvent,
    UISlider,
    UISpace,
    UISpriteWidget,
    UITextArea,
    UITextureButton,
    UITextureSlider,
    UITextureToggle,
    UIView,
)
from arcade.gui.experimental import UIPasswordInput

# Load system fonts
arcade.resources.load_kenney_fonts()

DEFAULT_FONT = ("Kenney Future", "arial")
DETAILS_FONT = ("arial", "Kenney Future Narrow")

# Preload textures, because they are mostly used multiple times, so they are not
# loaded multiple times
TEX_SCROLL_DOWN = arcade.load_texture(":resources:gui_basic_assets/scroll/indicator_down.png")
TEX_SCROLL_UP = arcade.load_texture(":resources:gui_basic_assets/scroll/indicator_up.png")

TEX_RED_BUTTON_NORMAL = arcade.load_texture(":resources:gui_basic_assets/button/red_normal.png")
TEX_RED_BUTTON_HOVER = arcade.load_texture(":resources:gui_basic_assets/button/red_hover.png")
TEX_RED_BUTTON_PRESS = arcade.load_texture(":resources:gui_basic_assets/button/red_press.png")
TEX_RED_BUTTON_DISABLE = arcade.load_texture(":resources:gui_basic_assets/button/red_disabled.png")

TEX_TOGGLE_RED = arcade.load_texture(":resources:gui_basic_assets/toggle/red.png")
TEX_TOGGLE_GREEN = arcade.load_texture(":resources:gui_basic_assets/toggle/green.png")

TEX_CHECKBOX_CHECKED = arcade.load_texture(":resources:gui_basic_assets/checkbox/blue_check.png")
TEX_CHECKBOX_UNCHECKED = arcade.load_texture(":resources:gui_basic_assets/checkbox/empty.png")

TEX_SLIDER_THUMB_BLUE = arcade.load_texture(":resources:gui_basic_assets/slider/thumb_blue.png")
TEX_SLIDER_TRACK_BLUE = arcade.load_texture(":resources:gui_basic_assets/slider/track_blue.png")
TEX_SLIDER_THUMB_RED = arcade.load_texture(":resources:gui_basic_assets/slider/thumb_red.png")
TEX_SLIDER_TRACK_RED = arcade.load_texture(":resources:gui_basic_assets/slider/track_red.png")
TEX_SLIDER_THUMB_GREEN = arcade.load_texture(":resources:gui_basic_assets/slider/thumb_green.png")
TEX_SLIDER_TRACK_GREEN = arcade.load_texture(":resources:gui_basic_assets/slider/track_green.png")

TEX_NINEPATCH_BASE = arcade.load_texture(":resources:gui_basic_assets/window/grey_panel.png")

TEX_ARCADE_LOGO = arcade.load_texture(":resources:/logo.png")

# Load animation for the sprite widget
frame_textures = []
for i in range(8):
    tex = arcade.load_texture(
        f":resources:images/animated_characters/female_adventurer/femaleAdventurer_walk{i}.png"
    )
    frame_textures.append(tex)

TEX_ANIMATED_CHARACTER = arcade.TextureAnimation(
    [arcade.TextureKeyframe(frame) for frame in frame_textures]
)

TEXT_WIDGET_EXPLANATION = textwrap.dedent("""
Arcade GUI provides three types of text widgets:


- {bold True}UILabel{bold False}:
A simple text widget that can be used to display text.

- {bold True}UIInputText{bold False}:
A text widget that can be used to get text input from the user.

- {bold True}UITextArea{bold False}:
A text widget that can be used to display text that is too long for a label.


This example shows how to use all three types of text widgets.


A few hints regarding the usage of the text widgets:


{bold True}UILabel{bold False}:

If you want to display frequently changing text,
setting a background color will improve performance.


{bold True}UIInputText{bold False}:

UIInputText dispatches an event on_change, when the text changes.


{bold True}UITextArea{bold False}:

While the widget supports scrolling, there is no scrollbar provided yet.
Users might oversee content.

In addition UITextArea supports different text styles,
which relate to Pyglet FormattedDocument.

"PLAIN" - Plain text.

"ATTRIBUTED" - Attributed text following the Pyglet attributed text style.

"HTML" - Allows to use HTML tags for formatting.

""").strip()


class ScrollableTextArea(UITextArea, UIAnchorLayout):
    """This widget is a text area that can be scrolled, like a UITextLayout, but shows indicator,
    that the text can be scrolled."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        indicator_size = 22
        self._down_indicator = UIImage(
            texture=TEX_SCROLL_DOWN, size_hint=None, width=indicator_size, height=indicator_size
        )
        self._down_indicator.visible = False
        self.add(self._down_indicator, anchor_x="right", anchor_y="bottom", align_x=3)

        self._up_indicator = UIImage(
            texture=TEX_SCROLL_UP, size_hint=None, width=indicator_size, height=indicator_size
        )
        self._up_indicator.visible = False
        self.add(self._up_indicator, anchor_x="right", anchor_y="top", align_x=3)

    def on_update(self, dt):
        self._up_indicator.visible = self.layout.view_y < 0
        self._down_indicator.visible = (
            abs(self.layout.view_y) < self.layout.content_height - self.layout.height
        )


class GalleryView(UIView):
    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_BELIZE_HOLE

        root = self.add_widget(UIAnchorLayout())

        # Setup side navigation
        nav_side = UIButtonRow(vertical=True, size_hint=(0.3, 1))
        nav_side.add(
            UILabel(
                "Categories",
                font_name=DEFAULT_FONT,
                font_size=32,
                text_color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE,
                size_hint=(1, 0.1),
                align="center",
            )
        )
        nav_side.add(UISpace(size_hint=(1, 0.01), color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE))

        nav_side.with_padding(all=10)
        nav_side.with_background(color=arcade.uicolor.WHITE_CLOUDS)
        nav_side.add_button("Start", style=UIFlatButton.STYLE_BLUE, size_hint=(1, 0.1))
        nav_side.add_button("Text", style=UIFlatButton.STYLE_BLUE, size_hint=(1, 0.1))
        nav_side.add_button("Interactive", style=UIFlatButton.STYLE_BLUE, size_hint=(1, 0.1))
        nav_side.add_button("Constructs", style=UIFlatButton.STYLE_BLUE, size_hint=(1, 0.1))
        nav_side.add_button("Other", style=UIFlatButton.STYLE_BLUE, size_hint=(1, 0.1))
        root.add(nav_side, anchor_x="left", anchor_y="top")

        @nav_side.event("on_action")
        def on_action(event: UIOnActionEvent):
            if event.action == "Start":
                self._show_start_widgets()
            elif event.action == "Text":
                self._show_text_widgets()
            elif event.action == "Interactive":
                self._show_interactive_widgets()
            elif event.action == "Constructs":
                self._show_construct_widgets()
            elif event.action == "Other":
                self._show_other_widgets()

        # Setup content to show widgets in

        self._body = UIAnchorLayout(size_hint=(0.7, 1))
        self._body.with_padding(all=20)
        root.add(self._body, anchor_x="right", anchor_y="top")

        # init start widgets
        self._show_start_widgets()

    def _show_start_widgets(self):
        """Show a short introduction message."""
        self._body.clear()
        self._body.add(
            UITextArea(
                text=textwrap.dedent("""
                Welcome to the Widget Gallery

                Choose a category on the left to see the widgets.
                You can checkout the source code in the examples/gui folder
                to see how this gallery is built.
                """).strip(),
                font_name=DETAILS_FONT,
                font_size=32,
                text_color=arcade.uicolor.WHITE,
                size_hint=(0.8, 0.8),
            ),
            anchor_y="top",
        )
        open_sourcecode = self._body.add(
            UIFlatButton(
                text="Open Source Code", style=UIFlatButton.STYLE_RED, size_hint=(0.3, 0.1)
            ),
            anchor_y="bottom",
            align_y=20,
        )

        @open_sourcecode.event("on_click")
        def on_click(_):
            """This will open the source code of this example on GitHub."""
            import webbrowser

            webbrowser.open(
                "https://github.com/pythonarcade/arcade/tree/development/arcade/examples/gui"
            )

    def _show_text_widgets(self):
        """Show the text widgets.

        For this we clear the body and add the text widgets.

        Featuring:
        - UILabel
        - UIInputText
        - UITextArea
        """

        self._body.clear()

        box = UIBoxLayout(vertical=True, size_hint=(1, 1), align="left", space_between=10)
        self._body.add(box)
        box.add(UILabel("Text Widgets", font_name=DEFAULT_FONT, font_size=32))
        box.add(UISpace(size_hint=(1, 0.1)))

        row_1 = UIBoxLayout(vertical=False, size_hint=(1, 0.1))
        box.add(row_1)
        row_1.add(UILabel("Username: ", font_name=DEFAULT_FONT, font_size=24))
        name_input = row_1.add(
            UIInputText(
                width=400,
                height=40,
                font_name=DEFAULT_FONT,
                font_size=24,
                border_color=arcade.uicolor.GRAY_CONCRETE,
                border_width=2,
            )
        )

        row_2 = UIBoxLayout(vertical=False, size_hint=(1, 0.1))
        box.add(row_2)
        row_2.add(UILabel("Password: ", font_name=DEFAULT_FONT, font_size=24))
        pw_input = row_2.add(
            UIPasswordInput(
                width=400,
                height=40,
                font_name=DEFAULT_FONT,
                font_size=24,
                border_color=arcade.uicolor.GRAY_CONCRETE,
                border_width=2,
            )
        )

        @pw_input.event("on_change")
        def _(event: UIOnChangeEvent):
            event.source.invalid = event.new_value != "arcade"

        welcome_label = box.add(
            UILabel("Nice to meet you ''", font_name=DEFAULT_FONT, font_size=24)
        )

        @name_input.event("on_change")
        def _(event: UIOnChangeEvent):
            welcome_label.text = f"Nice to meet you `{event.new_value}`"

        box.add(UISpace(size_hint=(1, 0.3)))  # Fill some of the left space

        text_area = box.add(
            ScrollableTextArea(
                text=TEXT_WIDGET_EXPLANATION,
                size_hint=(1, 0.9),
                font_name=DETAILS_FONT,
                font_size=16,
                text_color=arcade.uicolor.WHITE,
                document_mode="ATTRIBUTED",
            )
        )
        text_area.with_padding(left=10, right=10)
        text_area.with_border(color=arcade.uicolor.GRAY_CONCRETE, width=2)

    def _show_interactive_widgets(self):
        self._body.clear()
        box = UIBoxLayout(vertical=True, size_hint=(1, 1), align="left", space_between=10)
        self._body.add(box)
        box.add(UILabel("Interactive Widgets", font_name=DEFAULT_FONT, font_size=32))
        box.add(UISpace(size_hint=(1, 0.1)))

        flat_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(flat_row)

        flat_row.add(
            UIFlatButton(
                text="UIFlatButton blue", style=UIFlatButton.STYLE_BLUE, size_hint=(0.3, 1)
            )
        )
        flat_row.add(
            UIFlatButton(text="UIFlatButton red", style=UIFlatButton.STYLE_RED, size_hint=(0.3, 1))
        )
        flat_row.add(
            UIFlatButton(text="disabled", style=UIFlatButton.STYLE_BLUE, size_hint=(0.3, 1))
        ).disabled = True

        tex_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(tex_row)
        tex_row.add(
            UITextureButton(
                text="UITextureButton",
                texture=TEX_RED_BUTTON_NORMAL,
                texture_hovered=TEX_RED_BUTTON_HOVER,
                texture_pressed=TEX_RED_BUTTON_PRESS,
                texture_disabled=TEX_RED_BUTTON_DISABLE,
                size_hint=(0.3, 1),
            )
        )

        tex_row.add(UISpace(size_hint=(0.3, 1)))

        tex_row.add(
            UITextureButton(
                text="disabled",
                texture=TEX_RED_BUTTON_NORMAL,
                texture_hovered=TEX_RED_BUTTON_HOVER,
                texture_pressed=TEX_RED_BUTTON_PRESS,
                texture_disabled=TEX_RED_BUTTON_DISABLE,
                size_hint=(0.3, 1),
            )
        ).disabled = True

        toggle_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(toggle_row)
        toggle_row.add(
            UILabel("UITextureToggle", font_name=DETAILS_FONT, font_size=16, size_hint=(0.3, 0))
        )
        toggle_row.add(
            UITextureToggle(
                on_texture=TEX_TOGGLE_RED,
                off_texture=TEX_TOGGLE_GREEN,
                width=64,
                height=32,
            )
        )
        toggle_row.add(
            UITextureToggle(
                on_texture=TEX_CHECKBOX_CHECKED,
                off_texture=TEX_CHECKBOX_UNCHECKED,
                width=32,
                height=32,
            )
        )

        dropdown_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(dropdown_row)
        dropdown_row.add(
            UILabel("UIDropdown", font_name=DETAILS_FONT, font_size=16, size_hint=(0.3, 0))
        )
        dropdown_row.add(
            UIDropdown(
                default="Option 1",
                options=["Option 1", "Option 2", "Option 3"],
            )
        )

        slider_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(slider_row)

        slider_row.add(
            UILabel(
                "UISlider",
                font_name=DETAILS_FONT,
                font_size=16,
                size_hint=(0.3, 0),
            )
        )
        s1 = slider_row.add(UISlider(size_hint=(0.3, 1), step=1, style=UISlider.NO_STEP_STYLE))
        s2 = slider_row.add(
            UISlider(
                size_hint=(0.3, 1),
                step=5,
            )
        )
        s3 = slider_row.add(
            UISlider(
                size_hint=(0.3, 1),
                step=10,
            )
        )

        @s1.event("on_change")
        def _(event: UIOnChangeEvent):
            s2.value = event.new_value
            s3.value = event.new_value

        @s2.event("on_change")
        def _(event: UIOnChangeEvent):
            s1.value = event.new_value
            s3.value = event.new_value

        @s3.event("on_change")
        def _(event: UIOnChangeEvent):
            s1.value = event.new_value
            s2.value = event.new_value

        tex_slider_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(tex_slider_row)

        tex_slider_row.add(
            UILabel(
                "UITextureSlider",
                font_name=DETAILS_FONT,
                font_size=16,
                size_hint=(0.3, 0),
            )
        )

        ts1 = tex_slider_row.add(
            UITextureSlider(
                thumb_texture=TEX_SLIDER_THUMB_BLUE,
                track_texture=NinePatchTexture(10, 10, 10, 10, TEX_SLIDER_TRACK_BLUE),
                size_hint=(0.3, None),
            )
        )

        green_style = deepcopy(UITextureSlider.DEFAULT_STYLE)
        green_style["normal"].filled_track = arcade.uicolor.GREEN_GREEN_SEA
        green_style["hover"].filled_track = arcade.uicolor.GREEN_EMERALD
        green_style["press"].filled_track = arcade.uicolor.GREEN_GREEN_SEA
        ts2 = tex_slider_row.add(
            UITextureSlider(
                thumb_texture=TEX_SLIDER_THUMB_GREEN,
                track_texture=NinePatchTexture(10, 10, 10, 10, TEX_SLIDER_TRACK_GREEN),
                size_hint=(0.3, None),
                style=green_style,
            )
        )

        red_style = deepcopy(UITextureSlider.DEFAULT_STYLE)
        red_style["normal"].filled_track = arcade.uicolor.RED_POMEGRANATE
        red_style["hover"].filled_track = arcade.uicolor.RED_ALIZARIN
        red_style["press"].filled_track = arcade.uicolor.RED_POMEGRANATE
        ts3 = tex_slider_row.add(
            UITextureSlider(
                thumb_texture=TEX_SLIDER_THUMB_RED,
                track_texture=NinePatchTexture(10, 10, 10, 10, TEX_SLIDER_TRACK_RED),
                size_hint=(0.3, None),
                style=red_style,
            )
        )

        @ts1.event("on_change")
        def _(event: UIOnChangeEvent):
            ts2.value = event.new_value
            ts3.value = event.new_value

        @ts2.event("on_change")
        def _(event: UIOnChangeEvent):
            ts1.value = event.new_value
            ts3.value = event.new_value

        @ts3.event("on_change")
        def _(event: UIOnChangeEvent):
            ts1.value = event.new_value
            ts2.value = event.new_value

        box.add(UISpace(size_hint=(0.2, 0.1)))
        text_area = box.add(
            UITextArea(
                text=textwrap.dedent("""
                        Interactive widgets are widgets that the user can interact with.
                        This includes buttons, toggles, sliders and more.

                        By subclassing UIInteractiveWidget you
                        can create your own interactive widgets.

                        For text input have a look at the text widgets.
                    """).strip(),
                font_name=DETAILS_FONT,
                font_size=16,
                text_color=arcade.uicolor.WHITE,
                size_hint=(1, 0.9),
            )
        )
        text_area.with_padding(left=10, right=10)
        text_area.with_border(color=arcade.uicolor.GRAY_CONCRETE, width=2)

    def _show_construct_widgets(self):
        self._body.clear()
        box = UIBoxLayout(vertical=True, size_hint=(1, 1), align="left", space_between=10)
        self._body.add(box)
        box.add(UILabel("Constructs", font_name=DEFAULT_FONT, font_size=32))
        box.add(UISpace(size_hint=(1, 0.1)))

        message_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(message_row)
        message_row.add(
            UILabel(
                "UIMessageBox",
                font_name=DETAILS_FONT,
                font_size=16,
                size_hint=(0.3, 0),
            )
        )
        message_button = message_row.add(
            UIFlatButton(
                text="Show Message",
                style=UIFlatButton.STYLE_BLUE,
                size_hint=(0.3, 1),
            )
        )

        @message_button.event("on_click")
        def on_click(event):
            self.ui.add(
                UIMessageBox(
                    width=500,
                    height=350,
                    title="Message Box",
                    buttons=("Ok", "Cancel"),
                    message_text=textwrap.dedent("""
                    This is a message box.
                    It can be used to show messages to the user.

                    You can add buttons to it, to let the user choose an action.
                    """).strip(),
                ),
                layer=UIManager.OVERLAY_LAYER,
            )

        button_row = UIBoxLayout(vertical=False, size_hint=(1, 0.1), space_between=10)
        box.add(button_row)
        button_row.add(
            UILabel(
                "UIButtonRow",
                font_name=DETAILS_FONT,
                font_size=16,
                size_hint=(0.3, 0),
            )
        )
        buttons = button_row.add(
            UIButtonRow(
                text="Show Message",
                style=UIFlatButton.STYLE_BLUE,
                size_hint=(1, 0),
            )
        )
        buttons.add_button("Default Style", size_hint=(1, None))
        buttons.add_button("Red Style", style=UIFlatButton.STYLE_RED, size_hint=(1, None))
        buttons.add_button("Blue Style", style=UIFlatButton.STYLE_BLUE, size_hint=(1, None))

        # Constructs
        # "UIButtonRow",

        box.add(UISpace(size_hint=(0.2, 0.1)))
        text_area = box.add(
            UITextArea(
                text=textwrap.dedent("""
            Constructs are widgets that combine multiple widgets, to provide common functionality
            within a simple widget.
            Examples for this are message boxes or rows of buttons.
            """).strip(),
                font_name=DETAILS_FONT,
                font_size=16,
                text_color=arcade.uicolor.WHITE,
                size_hint=(1, 0.5),
            )
        )
        text_area.with_padding(left=10, right=10)
        text_area.with_border(color=arcade.uicolor.GRAY_CONCRETE, width=2)

    def _show_other_widgets(self):
        self._body.clear()
        box = UIBoxLayout(vertical=True, size_hint=(1, 1), align="left", space_between=10)
        self._body.add(box)
        box.add(UILabel("Other Widgets", font_name=DEFAULT_FONT, font_size=32))
        box.add(UISpace(size_hint=(1, 0.1)))

        image_row = box.add(UIBoxLayout(vertical=False, size_hint=(1, 0.1)))
        image_row.add(UILabel("UIImage", font_name=DETAILS_FONT, font_size=16, size_hint=(0.3, 0)))
        image_row.add(UIImage(texture=TEX_ARCADE_LOGO, width=64, height=64))

        dummy_row = box.add(UIBoxLayout(vertical=False, size_hint=(1, 0.1)))
        dummy_row.add(UILabel("UIDummy", font_name=DETAILS_FONT, font_size=16, size_hint=(0.3, 0)))
        dummy_row.add(UIDummy(size_hint=(0.2, 1)))
        dummy_row.add(UIDummy(size_hint=(0.2, 1)))
        dummy_row.add(UIDummy(size_hint=(0.2, 1)))

        sprite = arcade.TextureAnimationSprite(animation=TEX_ANIMATED_CHARACTER)
        sprite.scale = 0.5
        sprite_row = box.add(UIBoxLayout(vertical=False, size_hint=(1, 0.1)))
        sprite_row.add(
            UILabel("UISpriteWidget", font_name=DETAILS_FONT, font_size=16, size_hint=(0.3, 0))
        )
        sprite_row.add(UISpriteWidget(sprite=sprite, width=sprite.width, height=sprite.height))

        box.add(UISpace(size_hint=(0.2, 0.1)))
        text_area = box.add(
            UITextArea(
                text=textwrap.dedent("""
                    Arcade GUI provides also a few more widgets for special use cases.

                    - UIImage: A widget to display an image.
                    - UIDummy: Which can be used as a placeholder.
                                It renders a random color and changes it on click.
                    - UISpace: A widget that only takes up space.
                                But can also be used to add a colored space.
                    - UISpriteWidget: A widget that can display a sprite.

                    """).strip(),
                font_name=DETAILS_FONT,
                font_size=16,
                text_color=arcade.uicolor.WHITE,
                size_hint=(1, 0.9),
            )
        )
        text_area.with_padding(left=10, right=10)
        text_area.with_border(color=arcade.uicolor.GRAY_CONCRETE, width=2)


def main():
    window = arcade.Window(title="GUI Example: Widget Gallery")
    window.show_view(GalleryView())
    window.run()


if __name__ == "__main__":
    main()



"""
Texture Atlas for SpriteList

The long term goal is to rely on pyglet's texture atlas, but
it's still unclear what features we need supported in arcade
so need to prototype something to get started.

We're still building on pyglet's allocator.

Pyglet atlases are located here:
https://github.com/einarf/pyglet/blob/master/pyglet/image/atlas.py

Allocation:
Pyglet's allocator is a simple row based allocator only keeping
track of horizontal strips and how far in the x direction the
each strip is filled. We can't really "deallocate" unless it's
a region at the end of a strip and even doing that is awkward.

When an image is removed from the atlas we simply just lose that
region until we rebuild the atlas. It can be a good idea to count
the number of lost regions to use as an indicator later. When an
atlas is full we can first rebuild it if there are lost regions
instead of increasing the size.
"""

from __future__ import annotations

import abc
import contextlib
from pathlib import Path
from typing import TYPE_CHECKING

import PIL.Image

import arcade

if TYPE_CHECKING:
    from arcade import ArcadeContext, Texture
    from arcade.gl import Framebuffer, Texture2D
    from arcade.texture.texture import ImageData
    from arcade.texture_atlas.region import AtlasRegion

# The amount of pixels we increase the atlas when scanning for a reasonable size.
# It must be a power of two number like 64, 256, 512 ..
RESIZE_STEP = 128
UV_TEXTURE_WIDTH = 4096
# Texture coordinates for a texture (4 x vec2)
TexCoords = tuple[float, float, float, float, float, float, float, float]


class TextureAtlasBase(abc.ABC):
    """
    Abstract base class for texture atlases.

    A texture atlas is a large texture containing several textures
    so OpenGL can easily batch draw thousands or hundreds of thousands
    of sprites on one draw operation.
    """

    _fbo: Framebuffer
    _texture: Texture2D

    def __init__(self, ctx: ArcadeContext | None):
        self._ctx = ctx or arcade.get_window().ctx
        self._size: tuple[int, int] = 0, 0
        self._layers: int = 1

    @property
    def ctx(self) -> ArcadeContext:
        """The global Arcade OpenGL context."""
        return self._ctx

    @property
    def fbo(self) -> Framebuffer:
        """
        The framebuffer object for this atlas.

        This framebuffer has the atlas texture attached to it so
        we can render directly into the atlas texture.
        """
        return self._fbo

    @property
    def texture(self) -> Texture2D:
        """The OpenGL texture for this atlas."""
        return self._texture

    @property
    def width(self) -> int:
        """Hight of the atlas in pixels."""
        return self._size[0]

    @property
    def height(self) -> int:
        """Width of the atlas in pixels."""
        return self._size[1]

    @property
    def layers(self) -> int:
        """
        Number of layers in the atlas.
        Only relevant for atlases using texture arrays.
        """
        return self._layers

    @property
    def size(self) -> tuple[int, int]:
        """The width and height of the texture atlas in pixels"""
        return self._size

    # --- Core ---

    @abc.abstractmethod
    def add(self, texture: Texture) -> tuple[int, AtlasRegion]:
        """
        Add a texture to the atlas.

        Args:
            texture: The texture to add
        Returns:
            texture_id, AtlasRegion tuple
        Raises:
            AllocatorException: If there are no room for the texture
        """
        ...

    @abc.abstractmethod
    def remove(self, texture: Texture) -> None:
        """
        Remove a texture from the atlas.

        This is only supported by static atlases. Dynamic atlases with
        garbage collection will remove texture using python's garbage
        collector.

        Args:
            texture: The texture to remove
        """
        ...

    @abc.abstractmethod
    def resize(self, size: tuple[int, int], force=False) -> None:
        """
        Resize the atlas.

        This will re-allocate all the images in the atlas to better fit
        the new size. Pixel data will be copied from the old atlas to the
        new one on the gpu meaning it will also persist anything that
        was rendered to the atlas.

        A failed resize will result in an AllocatorException. Unless the
        atlas is resized again to a working size the atlas will be in an
        undefined state.

        Args:
            size:
                The new size
            force:
                Force a resize even if the size is the same
        """
        ...

    @abc.abstractmethod
    def rebuild(self) -> None:
        """
        Rebuild the underlying atlas texture.

        This method also tries to organize the textures more efficiently ordering them by size.
        The texture ids will persist so the sprite list doesn't need to be rebuilt.
        """
        ...

    @abc.abstractmethod
    def has_image(self, image_data: ImageData) -> bool:
        """
        Check if an image is already in the atlas

        Args:
            image_data: The image data to check
        """
        ...

    @abc.abstractmethod
    def has_texture(self, texture: Texture) -> bool:
        """
        Check if a texture is already in the atlas.

        Args:
            texture: The texture to check
        """
        ...

    @abc.abstractmethod
    def has_unique_texture(self, texture: Texture) -> bool:
        """
        Check if the atlas already have a texture with the
        same image data and vertex order

        Args:
            texture: The texture to check
        """
        ...

    @abc.abstractmethod
    def get_texture_id(self, texture: Texture) -> int:
        """
        Get the internal id for a Texture in the atlas

        Args:
            texture: The texture to get
        """
        ...

    @abc.abstractmethod
    def get_texture_region_info(self, atlas_name: str) -> AtlasRegion:
        """
        Get the region info for a texture by atlas name

        Args:
            atlas_name: The name of the texture in the atlas
        """
        ...

    @abc.abstractmethod
    def get_image_region_info(self, hash: str) -> AtlasRegion:
        """
        Get the region info for and image by has

        Args:
            hash: The hash of the image
        """
        ...

    @abc.abstractmethod
    def use_uv_texture(self, unit: int = 0) -> None:
        """
        Bind the texture coordinate texture to a channel.
        In addition this method writes the texture
        coordinate to the texture if the data is stale.
        This is to avoid a full update every time a texture
        is added to the atlas.

        Args:
            unit: The texture unit to bind the uv texture
        """
        ...

    # --- Utility ---

    @abc.abstractmethod
    @contextlib.contextmanager
    def render_into(
        self,
        texture: Texture,
        projection: tuple[float, float, float, float] | None = None,
    ):
        """
        Render directly into a sub-section of the atlas.
        The sub-section is defined by the already allocated space
        of the texture supplied in this method.

        By default the projection will be set to match the texture area size
        were `0, 0` is the lower left corner and `width, height` (of texture)
        is the upper right corner.

        This method should should be used with the ``with`` statement::

            with atlas.render_into(texture):
                # Draw commands here

            # Specify projection
            with atlas.render_into(texture, projection=(0, 100, 0, 100))
                # Draw geometry

        Args:
            texture:
                The texture area to render into
            projection:
                The ortho projection to render with. This parameter can be
                left blank if no projection changes are needed.
                The tuple values are: (left, right, button, top)
        """
        yield self._fbo

    @abc.abstractmethod
    def write_image(self, image: PIL.Image.Image, x: int, y: int) -> None:
        """
        Write a PIL image to the atlas in a specific region.

        Args:
            image:
                The pillow image
            x:
                The x position to write the texture
            y:
                The y position to write the texture
        """
        ...

    @abc.abstractmethod
    def read_texture_image_from_atlas(self, texture: Texture) -> PIL.Image.Image:
        """
        Read the pixel data for a texture directly from the atlas texture on the GPU.
        The contents of this image can be altered by rendering into the atlas and
        is useful in situations were you need the updated pixel data on the python side.

        Args:
            texture: The texture to get the image for
        Returns:
            A pillow image containing the pixel data in the atlas
        """
        ...

    @abc.abstractmethod
    def update_texture_image(self, texture: Texture):
        """
        Updates the internal image of a texture in the atlas texture.
        The new image needs to be the exact same size as the original
        one meaning the texture already need to exist in the atlas.

        This can be used in cases were the image is manipulated in some way
        and we need a quick way to sync these changes to graphics memory.
        This operation is fairly expensive, but still orders of magnitude
        faster than removing the old texture, adding the new one and
        re-building the entire atlas.

        Args:
            texture: The texture to update
        """
        ...

    @abc.abstractmethod
    def update_texture_image_from_atlas(self, texture: Texture) -> None:
        """
        Update the Arcade Texture's internal image with the pixel data content
        from the atlas texture on the GPU. This can be useful if you render
        into the atlas and need to update the texture with the new pixel data.

        Args:
            texture: The texture to update
        """
        ...

    # --- Debugging ---

    @abc.abstractmethod
    def to_image(
        self,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> PIL.Image.Image:
        """
        Convert the atlas to a Pillow image.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        Returns:
            A pillow image containing the atlas texture
        """
        ...

    @abc.abstractmethod
    def show(
        self,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> None:
        """
        Show the texture atlas using Pillow.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        """
        self.to_image(
            flip=flip,
            components=components,
            draw_borders=draw_borders,
            border_color=border_color,
        ).show()
        ...

    @abc.abstractmethod
    def save(
        self,
        path: str | Path,
        flip: bool = False,
        components: int = 4,
        draw_borders: bool = False,
        border_color: tuple[int, int, int] = (255, 0, 0),
    ) -> None:
        """
        Save the texture atlas to a png.

        Borders can also be drawn into the image to visualize the
        regions of the atlas.

        Args:
            path:
                The path to save the atlas on disk
            flip:
                Flip the image horizontally
            components:
                Number of components. (3 = RGB, 4 = RGBA)
            draw_borders:
                Draw region borders into image
            border_color:
                RGB color of the borders
        """


__all__ = (
    "TextureAtlasBase",
    "RESIZE_STEP",
    "UV_TEXTURE_WIDTH",
)



"""
Example showing how we can interact with spritelist using shaders.

We simply hijack the position buffer of the spritelist and
fill it with new data generated by a transform shader.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.spritelist_interaction_hijack_positions
"""

import math
import arcade
from arcade import hitbox

NUM_COINS = 500


class HijackSpritePositions(arcade.Window):

    def __init__(self):
        super().__init__(1280, 720, "Hijack Sprite Positions", resizable=True)
        # Generate lots of coins. We don't care about the initial positions
        # since our shader is setting those
        self.coins = arcade.SpriteList()
        texture = arcade.load_texture(
            ":resources:images/items/coinGold.png",
            hit_box_algorithm=hitbox.algo_bounding_box,
        )
        for _ in range(NUM_COINS):
            self.coins.append(arcade.Sprite(texture, scale=0.5))
        # Ensure the internal buffers are up to date (size etc)
        self.coins.write_sprite_buffers_to_gpu()

        # This shader simply generate some new positions
        self.position_program = self.ctx.program(
            vertex_shader="""
            #version 330

            // The current time to add some movement
            uniform float time;
            // The "bendyness" value accelerating rotations
            uniform float bend;

            // The current size of the screen
            uniform vec2 size;

            // The new positions we are writing into a new buffer
            out vec3 out_pos;

            void main() {
                // gl_VertexID is the sprite position in the spritelist.
                // We can use that to value to create unique positions with
                // some simple math.
                float vertId = float(gl_VertexID);
                out_pos = vec3(size, 0.0) / 2.0 + vec3(
                    sin(vertId + time + vertId * bend),
                    cos(vertId + time + vertId * bend),
                    0.0
                ) * vertId;
            }
            """
        )
        self.position_program["size"] = self.get_size()

    def on_draw(self):
        self.clear()

        # Write the new positions directly into the position
        # buffer of the spritelist. A little bit rude, but it works.
        self.coins.geometry.transform(
            self.position_program,
            self.coins.buffer_positions,
            vertices=len(self.coins),
        )
        self.coins.draw()

    def on_update(self, delta_time: float):
        # Keep updating the current time to animation the movement
        self.position_program["time"] = self.time / 4
        # Update the "bendyness" value
        self.position_program["bend"] = math.cos(self.time) / 400

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.position_program["size"] = width, height


HijackSpritePositions().run()



"""

from: https://github.com/linuxlewis/tripy/blob/master/tripy.py
"""

from arcade.types import Point2, Point2List


def earclip(
    polygon: Point2List,
) -> list[tuple[tuple[float, float], tuple[float, float], tuple[float, float]]]:
    """
    Simple earclipping algorithm for a given polygon p.

    polygon is expected to be an array of 2-tuples of the cartesian points of the polygon
    For a polygon with n points it will return n-2 triangles.
    The triangles are returned as an array of 3-tuples where each item in the tuple is a
    2-tuple of the cartesian point.

    Implementation Reference:
    https://www.geometrictools.com/Documentation/TriangulationByEarClipping.pdf

    Args:
        polygon: List of points in the polygon.
    Returns:
        A clipped polygon.
    """
    ear_vertex = []
    triangles = []

    polygon = list(polygon)

    if _is_clockwise(polygon):
        polygon.reverse()

    point_count = len(polygon)
    for i in range(point_count):
        prev_index = i - 1
        prev_point = polygon[prev_index]
        point = polygon[i]
        next_index = (i + 1) % point_count
        next_point = polygon[next_index]

        if _is_ear(prev_point, point, next_point, polygon):
            ear_vertex.append(point)

    while ear_vertex and point_count >= 3:
        ear = ear_vertex.pop(0)
        i = polygon.index(ear)
        prev_index = i - 1
        prev_point = polygon[prev_index]
        next_index = (i + 1) % point_count
        next_point = polygon[next_index]

        polygon.remove(ear)
        point_count -= 1
        triangles.append(
            ((prev_point[0], prev_point[1]), (ear[0], ear[1]), (next_point[0], next_point[1]))
        )
        if point_count > 3:
            prev_prev_point = polygon[prev_index - 1]
            next_next_index = (i + 1) % point_count
            next_next_point = polygon[next_next_index]

            groups = [
                (prev_prev_point, prev_point, next_point, polygon),
                (prev_point, next_point, next_next_point, polygon),
            ]
            for group in groups:
                p = group[1]
                if _is_ear(*group):
                    if p not in ear_vertex:
                        ear_vertex.append(p)
                elif p in ear_vertex:
                    ear_vertex.remove(p)
    return triangles


def _is_clockwise(polygon: Point2List) -> bool:
    """Check if the polygon is clockwise."""
    s = 0.0
    polygon_count = len(polygon)
    for i in range(polygon_count):
        point = polygon[i]
        point2 = polygon[(i + 1) % polygon_count]
        s += (point2[0] - point[0]) * (point2[1] + point[1])
    return s > 0


def _is_convex(prev: Point2, point: Point2, next_point: Point2) -> bool:
    """Check if the polygon is convex."""
    return _triangle_sum(prev[0], prev[1], point[0], point[1], next_point[0], next_point[1]) < 0


def _is_ear(p1: Point2, p2: Point2, p3: Point2, polygon: Point2List) -> bool:
    return (
        _contains_no_points(p1, p2, p3, polygon)
        and _is_convex(p1, p2, p3)
        and _triangle_area(p1[0], p1[1], p2[0], p2[1], p3[0], p3[1]) > 0
    )


def _contains_no_points(p1: Point2, p2: Point2, p3: Point2, polygon: Point2List) -> bool:
    for pn in polygon:
        if pn in (p1, p2, p3):
            continue
        elif _is_point_inside(pn, p1, p2, p3):
            return False
    return True


def _is_point_inside(p: Point2, a: Point2, b: Point2, c: Point2) -> bool:
    area = _triangle_area(a[0], a[1], b[0], b[1], c[0], c[1])
    area1 = _triangle_area(p[0], p[1], b[0], b[1], c[0], c[1])
    area2 = _triangle_area(p[0], p[1], a[0], a[1], c[0], c[1])
    area3 = _triangle_area(p[0], p[1], a[0], a[1], b[0], b[1])
    return area == sum([area1, area2, area3])


def _triangle_area(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float) -> float:
    return abs((x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)) / 2.0)


def _triangle_sum(x1: float, y1: float, x2: float, y2: float, x3: float, y3: float) -> float:
    return x1 * (y3 - y2) + x2 * (y1 - y3) + x3 * (y2 - y1)



"""
Array Backed Grid

Show how to use a two-dimensional list/array to back the display of a
grid on-screen.

Note: Regular drawing commands are slow. Particularly when drawing a lot of
items, like the rectangles in this example.

For faster drawing, create the shapes and then draw them as a batch.
See array_backed_grid_buffered.py

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.array_backed_grid
"""
import arcade

# Set how many rows and columns we will have
ROW_COUNT = 15
COLUMN_COUNT = 15

# This sets the WIDTH and HEIGHT of each grid location
WIDTH = 30
HEIGHT = 30

# This sets the margin between each cell
# and on the edges of the screen.
MARGIN = 5

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (WIDTH + MARGIN) * COLUMN_COUNT + MARGIN
WINDOW_HEIGHT = (HEIGHT + MARGIN) * ROW_COUNT + MARGIN
WINDOW_TITLE = "Array Backed Grid Example"


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """

        super().__init__()

        # Create a 2 dimensional array. A two-dimensional
        # array is simply a list of lists.
        self.grid = []
        for row in range(ROW_COUNT):
            # Add an empty array that will hold each cell
            # in this row
            self.grid.append([])
            for column in range(COLUMN_COUNT):
                self.grid[row].append(0)  # Append a cell

        self.background_color = arcade.color.BLACK

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Draw the grid
        for row in range(ROW_COUNT):
            for column in range(COLUMN_COUNT):
                # Figure out what color to draw the box
                if self.grid[row][column] == 1:
                    color = arcade.color.GREEN
                else:
                    color = arcade.color.WHITE

                # Do the math to figure out where the box is
                x = (MARGIN + WIDTH) * column + MARGIN + WIDTH // 2
                y = (MARGIN + HEIGHT) * row + MARGIN + HEIGHT // 2

                # Draw the box
                arcade.draw_rect_filled(arcade.rect.XYWH(x, y, WIDTH, HEIGHT), color)

    def on_mouse_press(self, x, y, button, modifiers):
        """
        Called when the user presses a mouse button.
        """

        # Change the x/y screen coordinates to grid coordinates
        column = int(x // (WIDTH + MARGIN))
        row = int(y // (HEIGHT + MARGIN))

        print(f"Click coordinates: ({x}, {y}). Grid coordinates: ({row}, {column})")

        # Make sure we are on-grid. It is possible to click in the upper right
        # corner in the margin and go to a grid location that doesn't exist
        if row < ROW_COUNT and column < COLUMN_COUNT:

            # Flip the location between 1 and 0.
            if self.grid[row][column] == 0:
                self.grid[row][column] = 1
            else:
                self.grid[row][column] = 0


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
THIS IS WORK IN PROGRESS. DO NOT USE.
"""

from typing import (
    TYPE_CHECKING,
)

from .base import TextureAtlasBase

if TYPE_CHECKING:
    from arcade import ArcadeContext


class TextureArrayAtlas(TextureAtlasBase):
    """
    A texture atlas that stores textures in a texture array.

    Args:
        ctx (ArcadeContext): The Arcade context.
        size (Tuple[int, int]): The texture size in pixels per layer
        layers (int): The number of layers (number of textures to store)
    """

    def __init__(self, ctx: ArcadeContext | None, size: tuple[int, int], layers: int):
        super().__init__(ctx)
        self._size = size
        self._layers = layers

    @property
    def size(self) -> tuple[int, int]:
        """The texture size in pixels per layer."""
        return self._size

    @property
    def layers(self) -> int:
        """The number of layers (number of textures to store)."""
        return self._layers



from PIL import Image, ImageDraw

import arcade
import arcade.cache
from arcade.types import Size2D

from .texture import ImageData, Texture

_DEFAULT_TEXTURE = None
_DEFAULT_IMAGE_SIZE = (128, 128)


def get_default_texture(size: Size2D[int] = _DEFAULT_IMAGE_SIZE) -> Texture:
    """
    Creates and returns a default texture and caches it internally for future use.

    Args:
        size: Size of the texture to create
    """
    global _DEFAULT_TEXTURE
    if _DEFAULT_TEXTURE:
        return _DEFAULT_TEXTURE

    _DEFAULT_TEXTURE = Texture(
        get_default_image(size),
        hit_box_algorithm=arcade.hitbox.algo_bounding_box,
    )
    return _DEFAULT_TEXTURE


def get_default_image(size: Size2D[int] = _DEFAULT_IMAGE_SIZE) -> ImageData:
    """
    Generates and returns a default image of the specified size
    and caches it in the default texture cache.

    Args:
        size: Size of the image
    """
    name = f"arcade-default-texture|{size}"
    image_data = arcade.texture.default_texture_cache.image_data_cache.get(name)
    if image_data:
        return image_data

    w, h = size
    im = Image.new("RGBA", (w, h), arcade.color.BLACK)
    draw = ImageDraw.Draw(im)
    draw.rectangle(((0, 0), (w // 2 - 1, h // 2 - 1)), arcade.color.MAGENTA)  # Upper left
    draw.rectangle(((w // 2, h // 2), (w, h)), arcade.color.MAGENTA)  # Lower right

    image_data = ImageData(im, hash=name)
    arcade.texture.default_texture_cache.image_data_cache.put(name, image_data)
    return image_data



"""
Example demonstrating controller support in an Arcade GUI.

This example shows how to integrate controller input with the Arcade GUI framework.
It includes a controller indicator widget that displays the last controller input,
and a modal dialog that can be navigated using a controller.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.exp_controller_support
"""


import arcade
from arcade import Texture
from arcade.experimental.controller_window import ControllerWindow, ControllerView
from arcade.gui import (
    UIAnchorLayout,
    UIBoxLayout,
    UIDropdown,
    UIEvent,
    UIFlatButton,
    UIImage,
    UIMouseFilterMixin,
    UIOnChangeEvent,
    UIOnClickEvent,
    UISlider,
    UIView,
)
from arcade.gui.events import (
    UIControllerButtonEvent,
    UIControllerButtonPressEvent,
    UIControllerDpadEvent,
    UIControllerEvent,
    UIControllerStickEvent,
    UIControllerTriggerEvent,
)
from arcade.gui.experimental.focus import UIFocusGroup
from arcade.types import Color


class ControllerIndicator(UIAnchorLayout):
    """
    A widget that displays the last controller input.
    """

    BLANK_TEX = Texture.create_empty("empty", (40, 40), arcade.color.TRANSPARENT_BLACK)
    TEXTURE_CACHE: dict[str, Texture] = {}

    def __init__(self):
        super().__init__()

        self._indicator = self.add(UIImage(texture=self.BLANK_TEX), anchor_y="bottom", align_y=10)
        self._indicator.with_background(color=Color(0, 0, 0, 0))
        self._indicator._strong_background = True

    @classmethod
    def get_texture(cls, path: str) -> Texture:
        if path not in cls.TEXTURE_CACHE:
            cls.TEXTURE_CACHE[path] = arcade.load_texture(path)
        return cls.TEXTURE_CACHE[path]

    @classmethod
    def input_prompts(cls, event: UIControllerEvent) -> Texture | None:
        if isinstance(event, UIControllerButtonEvent):
            match event.button:
                case "a":
                    return cls.get_texture(":resources:input_prompt/xbox/button_a.png")
                case "b":
                    return cls.get_texture(":resources:input_prompt/xbox/button_b.png")
                case "x":
                    return cls.get_texture(":resources:input_prompt/xbox/button_x.png")
                case "y":
                    return cls.get_texture(":resources:input_prompt/xbox/button_y.png")
                case "rightshoulder":
                    return cls.get_texture(":resources:input_prompt/xbox/rb.png")
                case "leftshoulder":
                    return cls.get_texture(":resources:input_prompt/xbox/lb.png")
                case "start":
                    return cls.get_texture(":resources:input_prompt/xbox/button_start.png")
                case "back":
                    return cls.get_texture(":resources:input_prompt/xbox/button_back.png")

        if isinstance(event, UIControllerTriggerEvent):
            match event.name:
                case "lefttrigger":
                    return cls.get_texture(":resources:input_prompt/xbox/lt.png")
                case "righttrigger":
                    return cls.get_texture(":resources:input_prompt/xbox/rt.png")

        if isinstance(event, UIControllerDpadEvent):
            match event.vector:
                case (1, 0):
                    return cls.get_texture(":resources:input_prompt/xbox/dpad_right.png")
                case (-1, 0):
                    return cls.get_texture(":resources:input_prompt/xbox/dpad_left.png")
                case (0, 1):
                    return cls.get_texture(":resources:input_prompt/xbox/dpad_up.png")
                case (0, -1):
                    return cls.get_texture(":resources:input_prompt/xbox/dpad_down.png")

        if isinstance(event, UIControllerStickEvent) and event.vector.length() > 0.2:
            stick = "l" if event.name == "leftstick" else "r"

            # map atan2(y, x) to direction string (up, down, left, right)
            heading = event.vector.heading()
            if 0.785 > heading > -0.785:
                return cls.get_texture(f":resources:input_prompt/xbox/stick_{stick}_right.png")
            elif 0.785 < heading < 2.356:
                return cls.get_texture(f":resources:input_prompt/xbox/stick_{stick}_up.png")
            elif heading > 2.356 or heading < -2.356:
                return cls.get_texture(f":resources:input_prompt/xbox/stick_{stick}_left.png")
            elif -2.356 < heading < -0.785:
                return cls.get_texture(f":resources:input_prompt/xbox/stick_{stick}_down.png")

        return None

    def on_event(self, event: UIEvent) -> bool | None:
        if isinstance(event, UIControllerEvent):
            input_texture = self.input_prompts(event)

            if input_texture:
                self._indicator.texture = input_texture

                arcade.unschedule(self.reset)
                arcade.schedule_once(self.reset, 0.5)

        return super().on_event(event)

    def reset(self, *_):
        self._indicator.texture = self.BLANK_TEX


class ControllerModal(UIMouseFilterMixin, UIFocusGroup):
    def __init__(self):
        super().__init__(size_hint=(0.8, 0.8))
        self.with_background(color=arcade.uicolor.DARK_BLUE_MIDNIGHT_BLUE)

        root = self.add(UIBoxLayout(space_between=10))

        root.add(UIFlatButton(text="Modal Button 1", width=200))
        root.add(UIFlatButton(text="Modal Button 2", width=200))
        root.add(UIFlatButton(text="Modal Button 3", width=200))
        root.add(UIFlatButton(text="Close")).on_click = self.close

        self.detect_focusable_widgets()

    def on_event(self, event):
        if super().on_event(event):
            return True

        if isinstance(event, UIControllerButtonPressEvent):
            if event.button == "b":
                self.close(None)
                return True

        return False

    def close(self, event):
        print("Close")
        # self.trigger_full_render()
        self.trigger_full_render()
        self.parent.remove(self)


class MyView(ControllerView, UIView):
    def __init__(self):
        super().__init__()
        arcade.set_background_color(arcade.color.AMAZON)

        base = self.add_widget(ControllerIndicator())
        self.root = base.add(UIFocusGroup())
        self.root.with_padding(left=10)
        box = self.root.add(UIBoxLayout(space_between=10), anchor_x="left")

        box.add(UIFlatButton(text="Button 1")).on_click = self.on_button_click
        box.add(UIFlatButton(text="Button 2")).on_click = self.on_button_click
        box.add(UIFlatButton(text="Button 3")).on_click = self.on_button_click

        box.add(UIDropdown(default="Option 1", options=["Option 1", "Option 2", "Option 3"]))

        slider = box.add(UISlider(value=0.5, min_value=0, max_value=1, width=200))

        @slider.event
        def on_change(event: UIOnChangeEvent):
            print(f"Slider value changed: {event}")

        self.root.detect_focusable_widgets()

    def on_button_click(self, event: UIOnClickEvent):
        print("Button clicked")
        self.root.add(ControllerModal())


if __name__ == "__main__":
    window = ControllerWindow(title="Controller UI Example")
    window.show_view(MyView())
    arcade.run()



"""
Various utility functions.

IMPORTANT: These should be standalone and not rely on any Arcade imports
"""

import platform
import sys
from collections.abc import Callable, Generator, Iterable, MutableSequence, Sequence
from itertools import chain
from pathlib import Path
from typing import Any, Generic, TypeVar

from arcade.types import AsFloat, Point2

__all__ = [
    "as_type",
    "type_name",
    "copy_dunders_unimplemented",
    "is_iterable",
    "is_nonstr_iterable",
    "is_str_or_noniterable",
    "grow_sequence",
    "is_raspberry_pi",
    "get_raspberry_pi_info",
]

# Since this module forbids importing from the rest of
# Arcade, we make our own local type variables.
_T = TypeVar("_T")
_TType = TypeVar("_TType", bound=type)


class Chain(Generic[_T]):
    """A reusable OOP version of :py:class:`itertools.chain`.

    In some cases (physics engines), we need to iterate over multiple
    sequences of objects repeatedly. This class provides a way to do so
    which:

    * Is non-exhausting
    * Avoids copying all items into one joined list

    Arguments:
        components: The sequences of items to join.
    """

    def __init__(self, *components: Sequence[_T]):
        self.components: list[Sequence[_T]] = list(components)

    def __iter__(self) -> Generator[_T, None, None]:
        yield from chain.from_iterable(self.components)


def as_type(item: Any) -> type:
    """If item is not a type, return its type. Otherwise, return item as-is.

    Args:
        item: A :py:class:`type` or instance of one.
    """
    if isinstance(item, type):
        return item
    else:
        return item.__class__


def type_name(item: Any) -> str:
    """Get the name of item if it's a type or the name of its type if it's an instance.

    This is meant to help shorten debugging-related code and developer
    utilities. It isn't meant to be a performant tool.

    Args:
         item: A :py:class:`type` or an instance of one.
    """
    if isinstance(item, type):
        return item.__name__
    else:
        return item.__class__.__name__


def is_iterable(item: Any) -> bool:
    """Use :py:func:`iter` to infer whether ``item`` is iterable.

    .. tip:: An empty iterable still counts as an iterable.

    This relies on the following:

    #. Python's :py:func:`iter` raises a :py:class:`TypeError` if
       ``item`` is not iterable
    #. try/catch blocks are fast

    If you are still concerned about performance, you may want to inline
    the contents of this function locally since function calls may have
    more overhead than try/catch blocks on your Python version.

    Args:
         item:
            An object to pass to the built-in :py:func:`iter` function.

    Returns:
        ``True`` if the item appears to be iterable.
    """
    try:
        _ = iter(item)
        return True
    except TypeError:
        return False


def is_nonstr_iterable(item: Any) -> bool:
    """``True`` if ``item`` is an iterable other than a :py:class:`str`.

    In addition to calling this function ``if`` and ``elif`` statements,
    you can also pass it as an argument to other functions. These include:

    * The :py:func:`.grow_sequence` utility function
    * Python's built-in filter function

    .. note:: This is the opposite of :py:func:`is_str_or_noniterable`.

    Args:
         item: Any object.

    Returns:
        Whether ``item`` is a non-string iterable.
    """
    return not isinstance(item, str) and is_iterable(item)


def is_str_or_noniterable(item: Any) -> bool:
    """``True`` if ``item`` is a string or non-iterable.

    In addition to calling this function in ``if`` and ``elif`` statements,
    you can also pass it as an argument to other functions. These include:

    * The :py:func:`.grow_sequence` utility function
    * Python's built-in :py:func:`filter` function

    .. note:: This is the opposite of :py:func:`is_str_or_noniterable`.

    Args:
        item: Any object.

    Returns:
        ``True`` if ``item`` is a :py:class:`str` or a non-iterable
        object.
    """
    return isinstance(item, str) or not is_iterable(item)


def grow_sequence(
    destination: MutableSequence[_T],
    source: _T | Iterable[_T],
    append_if: Callable[[_T | Iterable[_T]], bool] = is_str_or_noniterable,
) -> None:
    """Append when ``append_if(to_add)`` is ``True``, extend otherwise.

    If performance is critical, consider inlining this code. This function
    is meant as:

    * a companion to :py:data:`arcade.types.OneOrIterableOf`
    * an abbreviation for repetitive if-blocks in config and settings menus

    The default ``append_if`` value is the :py:func:`.is_str_or_noniterable`
    function in this module. You can pass any :py:func:`~typing.Callable`
    which returns:

    * ``True`` if we should :py:meth:`append <list.append>`
    * ``False`` if we should :py:meth:`extend <list.extend>`

    This includes both the :py:class:`callable` function and your own custom
    functions. For example:

    .. code-block:: python

       dest = []

       def _validate_and_choose(item) -> bool:
           \"\"\"Raises an exception if data is invalid or a bool if not.\"\"\"
           ...

       # Okay values
       grow_sequence(dest, MyType(), append_when=_validate_and_choose)
       grow_sequence(dest, [MyType(), MyType()], append_when=_validate_and_choose)

       # Raises an exception
       grow_sequence(dest, BadType(), append_when=_validate_and_choose)


    Args:
        destination:
            A :py:func:`list` or other :py:class:`~typing.MutableSequence`
            to append to or extend.
        source:
            A value source we'll use to grow the ``destination``
            sequence.
        append_if:
            A :py:func:`callable` which returns ``True`` when ``source``
            should be appended.
    """
    if append_if(source):
        destination.append(source)  # type: ignore
    else:
        destination.extend(source)  # type: ignore


def copy_dunders_unimplemented(decorated_type: _TType) -> _TType:
    """Decorator stubs dunders raising :py:class:`NotImplementedError`.

    Temp fixes https://github.com/pythonarcade/arcade/issues/2074 by
    stubbing the following instance methods:

    * :py:meth:`object.__copy__` (used by :py:func:`copy.copy`)
    * :py:meth:`object.__deepcopy__` (used by :py:func:`copy.deepcopy`)

    Example usage:

    .. code-block:: python

       import copy
       from arcade,utils import copy_dunders_unimplemented
       from arcade.hypothetical_module import HypotheticalNasty

       # Example usage
       @copy_dunders_unimplemented
       class CantCopy:
            def __init__(self, nasty_state: HypotheticalNasty):
                self.nasty_state = nasty_state

       instance = CantCopy(HypotheticalNasty())

       # These raise NotImplementedError
       this_line_raises = copy.deepcopy(instance)
       this_line_also_raises = copy.copy(instance)
    """

    def __copy__(self):  # noqa
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement __copy__, but"
            f"you may implement it on a custom subclass."
        )

    decorated_type.__copy__ = __copy__  # type: ignore

    def __deepcopy__(self, memo):  # noqa
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement __deepcopy__,"
            f" but you may implement it on a custom subclass."
        )

    decorated_type.__deepcopy__ = __deepcopy__  # type: ignore

    return decorated_type


def is_raspberry_pi() -> bool:
    """Determine if the host is a raspberry pi."""
    return get_raspberry_pi_info()[0]


def get_raspberry_pi_info() -> tuple[bool, str, str]:
    """
    Determine if the host is a raspberry pi with additional info.

    Returned tuple format is::

            bool (is host a raspi)
            str (architecture)
            str (model name)
    """
    # The platform for raspi should always be linux
    if not sys.platform == "linux":
        return False, "", ""

    # armv7l is raspi 32 bit
    # aarch64 is raspi 64 bit
    architecture = platform.machine()
    model_name = ""

    # Check for model info file
    MODEL_PATH = Path("/sys/firmware/devicetree/base/model")
    if MODEL_PATH.exists():
        try:
            model_name = MODEL_PATH.read_text()[:-1]
            if "raspberry pi" in model_name.lower():
                return True, architecture, model_name
        except Exception:
            pass

    return False, "", ""


def unpack_asfloat_or_point(value: AsFloat | Point2) -> Point2:
    """
    A utility method that converts a float or int into a Point2, or
    validates that an iterable is a Point2.

    .. note:: This should be inlined in hot code paths

    Args:
        value: The value to test.

    Returns:
        A Point2 that is either equal to value, or is equal to (value, value)
    """

    if isinstance(value, float | int):
        x = y = value
    else:
        try:
            x, y = value
        except ValueError:
            raise ValueError(
                "value must be a float, int, or tuple-like which unpacks as two float-like values"
            )
        except TypeError:
            raise TypeError(
                "value must be a float, int, or tuple-like unpacks as two float-like values"
            )

    return x, y



"""
Sprite Health Bars

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_health
"""
import math

import arcade
from arcade.types import Color

IMAGE_FEMALE_PERSON_IDLE = (
    ":assets:images/animated_characters/female_person/femalePerson_idle.png"
)
IMAGE_LASER_BLUE01 = ":assets:images/space_shooter/laserBlue01.png"
IMAGE_ZOMBIE_IDLE = ":assets:images/animated_characters/zombie/zombie_idle.png"

SPRITE_SCALING_PLAYER = 0.5
SPRITE_SCALING_ENEMY = 0.5
SPRITE_SCALING_BULLET = 1
INDICATOR_BAR_OFFSET = 32
ENEMY_ATTACK_COOLDOWN = 1
BULLET_SPEED = 150
BULLET_DAMAGE = 1
PLAYER_HEALTH = 5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Health Bars"


def sprite_off_screen(
    sprite: arcade.Sprite,
    WINDOW_HEIGHT: int = WINDOW_HEIGHT,
    WINDOW_WIDTH: int = WINDOW_WIDTH,
) -> bool:
    """Checks if a sprite is off-screen or not."""
    return (
        sprite.top < 0
        or sprite.bottom > WINDOW_HEIGHT
        or sprite.right < 0
        or sprite.left > WINDOW_WIDTH
    )


class Player(arcade.Sprite):
    def __init__(self, bar_list: arcade.SpriteList) -> None:
        super().__init__(
            IMAGE_FEMALE_PERSON_IDLE,
            scale=SPRITE_SCALING_PLAYER,
        )
        self.indicator_bar: IndicatorBar = IndicatorBar(
            self, bar_list, (self.center_x, self.center_y), scale=(1.5, 1.5),
        )
        self.health: int = PLAYER_HEALTH


class Bullet(arcade.Sprite):
    def __init__(self) -> None:
        super().__init__(
            IMAGE_LASER_BLUE01,
            scale=SPRITE_SCALING_BULLET,
        )

    def update(self, delta_time: float = 1 / 60) -> None:
        """Updates the bullet's position."""
        self.position = (
            self.center_x + self.change_x * delta_time,
            self.center_y + self.change_y * delta_time,
        )


class IndicatorBar:
    """
    Represents a bar which can display information about a sprite.

    Args:
        owner:
            The owner of this indicator bar.
        sprite_list:
            The sprite list used to draw the indicator bar components.
        position:
            The initial position of the bar.
        full_color:
            The color of the bar.
        background_color:
            The background color of the bar.
        width:
            The width of the bar.
        height:
            The height of the bar.
        border_size:
            The size of the bar's border.
        scale:
            The scale of the indicator bar.
    """

    def __init__(
        self,
        owner: Player,
        sprite_list: arcade.SpriteList,
        position: tuple[float, float] = (0, 0),
        full_color: Color = arcade.color.GREEN,
        background_color: Color = arcade.color.BLACK,
        width: int = 100,
        height: int = 4,
        border_size: int = 4,
        scale: tuple[float, float] = (1.0, 1.0),
    ) -> None:
        # Store the reference to the owner and the sprite list
        self.owner: Player = owner
        self.sprite_list: arcade.SpriteList = sprite_list

        # Set the needed size variables
        self._bar_width: int = width
        self._bar_height: int = height
        self._center_x: float = 0.0
        self._center_y: float = 0.0
        self._fullness: float = 0.0
        self._scale: tuple[float, float] = (1.0, 1.0)

        # Create the boxes needed to represent the indicator bar
        self._background_box: arcade.SpriteSolidColor = arcade.SpriteSolidColor(
            self._bar_width + border_size,
            self._bar_height + border_size,
            color=background_color,
        )
        self._full_box: arcade.SpriteSolidColor = arcade.SpriteSolidColor(
            self._bar_width,
            self._bar_height,
            color=full_color,
        )
        self.sprite_list.append(self._background_box)
        self.sprite_list.append(self._full_box)

        # Set the fullness, position and scale of the bar
        self.fullness = 1.0
        self.position = position
        self.scale = scale

    def __repr__(self) -> str:
        return f"<IndicatorBar (Owner={self.owner})>"

    @property
    def background_box(self) -> arcade.SpriteSolidColor:
        """Returns the background box of the indicator bar."""
        return self._background_box

    @property
    def full_box(self) -> arcade.SpriteSolidColor:
        """Returns the full box of the indicator bar."""
        return self._full_box

    @property
    def bar_width(self) -> int:
        """Gets the width of the bar."""
        return self._bar_width

    @property
    def bar_height(self) -> int:
        """Gets the height of the bar."""
        return self._bar_height

    @property
    def center_x(self) -> float:
        """Gets the x position of the bar."""
        return self._center_x

    @property
    def center_y(self) -> float:
        """Gets the y position of the bar."""
        return self._center_y

    @property
    def top(self) -> float:
        """Gets the y coordinate of the top of the bar."""
        return self.background_box.top

    @property
    def bottom(self) -> float:
        """Gets the y coordinate of the bottom of the bar."""
        return self.background_box.bottom

    @property
    def left(self) -> float:
        """Gets the x coordinate of the left of the bar."""
        return self.background_box.left

    @property
    def right(self) -> float:
        """Gets the x coordinate of the right of the bar."""
        return self.background_box.right

    @property
    def fullness(self) -> float:
        """Returns the fullness of the bar."""
        return self._fullness

    @fullness.setter
    def fullness(self, new_fullness: float) -> None:
        """Sets the fullness of the bar."""
        # Check if new_fullness if valid
        if not (0.0 <= new_fullness <= 1.0):
            raise ValueError(
                f"Got {new_fullness}, but fullness must be between 0.0 and 1.0."
            )

        # Set the size of the bar
        self._fullness = new_fullness
        if new_fullness == 0.0:
            # Set the full_box to not be visible since it is not full anymore
            self.full_box.visible = False
        else:
            # Set the full_box to be visible incase it wasn't then update the bar
            self.full_box.visible = True
            self.full_box.width = self._bar_width * new_fullness * self.scale[0]
            self.full_box.left = self._center_x - (self._bar_width / 2) * self.scale[0]

    @property
    def position(self) -> tuple[float, float]:
        """Returns the current position of the bar."""
        return self._center_x, self._center_y

    @position.setter
    def position(self, new_position: tuple[float, float]) -> None:
        """Sets the new position of the bar."""
        # Check if the position has changed. If so, change the bar's position
        if new_position != self.position:
            self._center_x, self._center_y = new_position
            self.background_box.position = new_position
            self.full_box.position = new_position

            # Make sure full_box is to the left of the bar instead of the middle
            self.full_box.left = self._center_x - (self._bar_width / 2) * self.scale[0]

    @property
    def scale(self) -> tuple[float, float]:
        """Returns the scale of the bar."""
        return self._scale

    @scale.setter
    def scale(self, value: tuple[float, float]) -> None:
        """Sets the new scale of the bar."""
        # Check if the scale has changed. If so, change the bar's scale
        if value != self.scale:
            self._scale = value
            self.background_box.scale = value
            self.full_box.scale = value


class GameView(arcade.View):
    def __init__(self) -> None:
        super().__init__()

        # Create sprite lists
        self.bullet_list: arcade.SpriteList = arcade.SpriteList()
        self.bar_list: arcade.SpriteList = arcade.SpriteList()
        self.player_sprite_list: arcade.SpriteList = arcade.SpriteList()
        self.enemy_sprite_list: arcade.SpriteList = arcade.SpriteList()

        # Create player sprite
        self.player_sprite = Player(self.bar_list)
        self.player_sprite_list.append(self.player_sprite)

        # Create enemy Sprite
        self.enemy_sprite = arcade.Sprite(IMAGE_ZOMBIE_IDLE, scale=SPRITE_SCALING_ENEMY)
        self.enemy_sprite_list.append(self.enemy_sprite)

        # Create text objects
        self.top_text: arcade.Text = arcade.Text(
            "Dodge the bullets by moving the mouse!",
            self.width // 2,
            self.height - 50,
            anchor_x="center",
        )
        self.bottom_text: arcade.Text = arcade.Text(
            "When your health bar reaches zero, you lose!",
            self.width // 2,
            50,
            anchor_x="center",
        )
        self.enemy_timer = 0

    def setup(self) -> None:
        """Set up the game and initialize the variables."""
        # Setup player and enemy positions
        self.player_sprite.position = self.width // 2, self.height // 4
        self.enemy_sprite.position = self.width // 2, self.height // 2

        # Set the background color
        self.background_color = arcade.color.AMAZON

    def on_draw(self) -> None:
        """Render the screen."""
        # Clear the screen. This command has to happen before we start drawing
        self.clear()

        # Draw all the sprites
        self.player_sprite_list.draw()
        self.enemy_sprite_list.draw()
        self.bullet_list.draw()
        self.bar_list.draw()

        # Draw the text objects
        self.top_text.draw()
        self.bottom_text.draw()

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float) -> None:
        """Called whenever the mouse moves."""
        self.player_sprite.position = x, y

    def on_update(self, delta_time) -> None:
        """Movement and game logic."""
        # Check if the player is dead. If so, exit the game
        if self.player_sprite.health <= 0:
            arcade.exit()

        # Increase the enemy's timer
        self.enemy_timer += delta_time

        # Update the player's indicator bar position
        self.player_sprite.indicator_bar.position = (
            self.player_sprite.center_x,
            self.player_sprite.center_y + INDICATOR_BAR_OFFSET,
        )

        # Call updates on bullet sprites
        self.bullet_list.update(delta_time)

        # Check if the enemy can attack. If so, shoot a bullet from the
        # enemy towards the player
        if self.enemy_timer >= ENEMY_ATTACK_COOLDOWN:
            self.enemy_timer = 0

            # Create the bullet
            bullet = Bullet()

            # Set the bullet's position
            bullet.position = self.enemy_sprite.position

            # Calculate the trajectory.
            # Zero degrees is up, 90 to the right.
            # atan returns 0 degrees to the right instead of up, so shift by 90 degrees.
            diff_x = self.player_sprite.center_x - self.enemy_sprite.center_x
            diff_y = self.player_sprite.center_y - self.enemy_sprite.center_y
            angle = -math.atan2(diff_y, diff_x) + 3.14 / 2
            angle_deg = math.degrees(angle)

            if angle_deg < 0:
                angle_deg += 360

            # Set the bullet's angle to face the player.
            # Bullet graphic isn't pointed up, so rotate 90
            bullet.angle = angle_deg - 90

            # Give the bullet a velocity towards the player
            bullet.change_x = math.sin(angle) * BULLET_SPEED
            bullet.change_y = math.cos(angle) * BULLET_SPEED

            # Add the bullet to the bullet list
            self.bullet_list.append(bullet)

        # Loop through each bullet
        for existing_bullet in self.bullet_list:
            # Check if the bullet has gone off-screen. If so, delete the bullet
            if sprite_off_screen(existing_bullet):
                existing_bullet.remove_from_sprite_lists()
                continue

            # Check if the bullet has hit the player
            if arcade.check_for_collision(existing_bullet, self.player_sprite):
                # Damage the player and remove the bullet
                self.player_sprite.health -= BULLET_DAMAGE
                existing_bullet.remove_from_sprite_lists()

                # Set the player's indicator bar fullness
                self.player_sprite.indicator_bar.fullness = (
                    self.player_sprite.health / PLAYER_HEALTH
                )


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Shows how we can use shaders using existing spritelist data.

This examples renders a line between the player position
and nearby sprites when they are within a certain distance.

This builds on a previous example adding line of sight (LoS)
checks by using texture lookups. We our walls into a
texture and read the pixels in a line between the
player and the target sprite to check if the path is
colliding with something.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.spritelist_interaction_visualize_dist_los
"""

import random
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
NUM_COINS = 500
NUM_WALLS = 75
INTERACTION_RADIUS = 300


class SpriteListInteraction(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, "SpriteList Interaction - LoS")

        # Player
        self.player = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.25,
        )

        # Wall sprites we are checking collision against
        self.walls = arcade.SpriteList()
        for _ in range(NUM_WALLS):
            self.walls.append(
                arcade.Sprite(
                    ":resources:images/tiles/boxCrate_double.png",
                    center_x=random.randint(0, WINDOW_WIDTH),
                    center_y=random.randint(0, WINDOW_HEIGHT),
                    scale=0.25,
                )
            )

        # Generate some random coins.
        # We make sure they are not placed inside a wall.
        # We give the coins one chance to appear outside walls
        self.coins = arcade.SpriteList()
        for _ in range(NUM_COINS):
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                center_x=random.randint(0, WINDOW_WIDTH),
                center_y=random.randint(0, WINDOW_HEIGHT),
                scale=0.25,
            )
            if arcade.check_for_collision_with_list(coin, self.walls):
                continue

            self.coins.append(coin)

        # This program draws lines from the player/origin
        # to sprites that are within a certain distance.
        # The main action here happens in the geometry shader.
        # It creates lines when a sprite is within the maxDistance.
        self.program_visualize_dist = self.ctx.program(
            vertex_shader="""
            #version 330

            // Sprite positions from SpriteList
            in vec3 in_pos;

            // Output to geometry shader
            out vec3 v_position;

            void main() {
                // This shader just forwards info to geo shader
                v_position = in_pos;
            }
            """,
            geometry_shader="""
            #version 330

            // This is how we access pyglet's global projection matrix
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            // The position we measure distance from
            uniform vec2 origin;
            // The maximum distance
            uniform float maxDistance;
            // Sampler for reading wall data
            uniform sampler2D walls;

            // These configure the geometry shader to process a points
            // and allows it to emit lines. It runs for every sprite
            // in the spritelist.
            layout (points) in;
            layout (line_strip, max_vertices = 2) out;

            // The position input from vertex shader.
            // It's an array because geo shader can take more than one input
            in vec3 v_position[];

            // Helper function converting screen coordinates to texture coordinates.
            // Texture coordinates are normalized (0.0 -> 1.0) were 0,0 is in the
            vec2 screen2texcoord(vec2 pos) {
                return vec2(pos / vec2(textureSize(walls, 0).xy));
            }

            void main() {
                // ONLY emit a line between the sprite and origin when within the distance
                if (distance(v_position[0].xy, origin) > maxDistance) return;

                // Read samples from the wall texture in a line looking for obstacles
                // We simply make a vector between the original and the sprite location
                // and trace pixels in this path with a reasonable step.
                int numSteps = int(maxDistance / 2.0);
                vec2 dir = v_position[0].xy - origin;
                for (int i = 0; i < numSteps; i++) {
                    // Read pixels along the vector
                    vec2 pos = origin + dir * (float(i) / float(numSteps));
                    vec4 color = texture(walls, screen2texcoord(pos));
                    // If we find non-zero pixel data we have obstacles in our path!
                    if (color != vec4(0.0)) return;
                }

                // First line segment position (origin)
                gl_Position = window.projection * window.view * vec4(origin, 0.0, 1.0);
                EmitVertex();
                // Second line segment position (sprite position)
                gl_Position = window.projection * window.view * vec4(v_position[0].xy, 0.0, 1.0);
                EmitVertex();
                EndPrimitive();
            }
            """,
            fragment_shader="""
            #version 330
            // The fragment shader just runs for every pixel of the line segment.

            // Reference to the pixel we are writing to in the framebuffer
            out vec4 fragColor;

            void main() {
                // All the pixels in the line should just be white
                fragColor = vec4(1.0, 1.0, 1.0, 1.0);
            }
            """,
        )
        # Configure program with maximum distance
        self.program_visualize_dist["maxDistance"] = INTERACTION_RADIUS

        # Lookup texture/framebuffer for walls so we can trace pixels in the shader.
        # It contains a texture attachment with the same size as the window.
        # We draw only the walls into this one as a line of sight lookup
        self.walls_fbo = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture((WINDOW_WIDTH, WINDOW_HEIGHT))]
        )
        # Draw the walls into the framebuffer
        with self.walls_fbo.activate() as fbo:
            fbo.clear()
            self.walls.draw()

    def on_draw(self):
        self.clear()

        self.walls.draw()
        self.coins.draw()
        # Bind the wall texture to texture channel 0 so we can read it in the shader
        self.walls_fbo.color_attachments[0].use(0)
        # We already have a geometry instance in the spritelist we can
        # use to run our shader/gpu program. It only requires that we
        # use correctly named input name(s). in_pos in this example
        # what will automatically map in the position buffer to the vertex shader.
        self.coins.geometry.render(self.program_visualize_dist, vertices=len(self.coins))
        arcade.draw_sprite(self.player)

        # Visualize the interaction radius
        arcade.draw_circle_filled(
            center_x=self.player.center_x,
            center_y=self.player.center_y,
            radius=INTERACTION_RADIUS,
            color=(255, 255, 255, 64)
        )

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float):
        # Move the sprite to mouse position
        self.player.position = x, y
        # Update the program with a new origin
        self.program_visualize_dist["origin"] = x, y


SpriteListInteraction().run()



"""
Compute shader with buffers.

The example is using compute shaders with shader storage buffers (SSBOs).
These are simply large buffers we can read and write to. We also show
how we can visualize the result using a normal shader.

We generate N number of points with random velocities, colors and a radius in python
and write that to an ssbo/buffer. Every frame we run a compute shader
that calculates new positions for every point based on the velocities,
collision with screen borders and a directional force we change over time.
The result ends up in a new buffer because it's not always practical to
read and write to the same buffer at the same time (but it is possible).
We swap the source and target buffer every frame so we keep working on
the previous state (ping-pong).

These points are then rendered as circles/balls using a traditional shader.
We use a geometry shader to on-the-fly create a rectangle for each point.
The size of each ball is based on the radius stored for each point.

The compute shader in this example works in one dimension for simplicity.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.compute_ssbo
"""

import random
import math
from array import array

import arcade
from arcade.gl import BufferDescription

COMPUTE_SHADER = """
#version 430

// Describe the minimum amount of work this compute shader can execute.
// This is called a work group. Our work group will run 256 threads/invocations.
// When we run a compute shader we specify how many of these groups we are executing.
// This work group only has one dimension, but there is also local_size_y and local_size_z.
layout(local_size_x=256) in;

// Global variables for this shader.
// These are assigned to the shader at runtime.
uniform vec2 screen_size;
uniform vec2 force;
uniform float frame_time;

// Use a structure to make it easier to describe the input data.
// Note that these vectors are larger than they need to be
// to provide the compute shader with neatly aligned data.
struct Ball
{
    vec4 pos;
    vec4 vel;
    vec4 col;
};

// Input buffer of points/balls
layout(std430, binding=0) buffer balls_in
{
    Ball balls[];
} In;
// Output buffer of points/balls
layout(std430, binding=1) buffer balls_out
{
    Ball balls[];
} Out;

void main()
{
    // Get the ball number we are working with.
    // This goes from 0 to N (number of balls)
    int x = int(gl_GlobalInvocationID.x);

    // Get the current ball from the input buffer
    Ball in_ball = In.balls[x];
    // Get position for easier access
    vec4 p = in_ball.pos.xyzw;
    // Get the velocity for easy access
    vec4 v = in_ball.vel.xyzw;

    // Apply velocity
    p.xy += v.xy;

    // Apply external force vector
    v.xy += force * frame_time;

    // Get at weak the radius
    float rad = p.w * 0.5;

    // Check If we are outside the screen.
    // Tweak position and velocity it this is the case.
    // The ball will lose a bit of velocity on collisions.
    if (p.x - rad <= 0.0)
    {
        p.x = rad;
        v.x *= -0.98;
    }
    else if (p.x + rad >= screen_size.x)
    {
        p.x = screen_size.x - rad;
        v.x *= -0.98;
    }

    if (p.y - rad <= 0.0)
    {
        p.y = rad;
        v.y *= -0.98;
    }
    else if (p.y + rad >= screen_size.y)
    {
        p.y = screen_size.y - rad;
        v.y *= -0.98;
    }

    // Create new ball and write to out buffer
    Ball out_ball;
    out_ball.pos = p;
    out_ball.vel = v;
    out_ball.col = in_ball.col;

    Out.balls[x] = out_ball;
}
"""


class App(arcade.Window):

    def __init__(self, *args, **kwargs):
        # We need to specify OpenGL 4.3 when using Compute Shaders
        super().__init__(
            1280, 720, "Compute Shader",
            gl_version=(4, 3), resizable=True, vsync=True,
        )
        # Keep track of time
        self.frame_time = 0
        # The work group size we have configured the compute shader to.
        # This is hardcoded the compute shader (careful with changing)
        self.work_group_size = 256
        # How many work groups to execute when running the compute shader
        self.work_groups_to_execute = 20
        # How many balls/points to work with
        self.num_balls = self.work_group_size * self.work_groups_to_execute

        # Create the compute shader. We only need the source.
        self.cs = self.ctx.compute_shader(source=COMPUTE_SHADER)

        # Buffers with data for the compute shader (We ping-pong render between these).
        # The generated data is 12 x 32bit floats containing positions, velocities and
        # colors. Some of these floats are just dummy/padding floats to make the data
        # neatly aligned for the compute shader. There are strict rules for this
        # for a compute shader to be able to parallelize the work over potentially
        # thousands of threads.
        self.ssbo_1 = self.ctx.buffer(data=array("f", self.gen_initial_data(self.num_balls)))
        # Create a second buffer with the exact same byte size
        self.ssbo_2 = self.ctx.buffer(reserve=self.ssbo_1.size)

        # Create a geometry instance describing the buffer contents so the
        # shader visualizing them knows how to map the buffer data to vertex shader inputs.
        # We need one for each buffer because the data is moving between the
        # two buffers every frame.
        self.vao_1 = self.ctx.geometry(
            [BufferDescription(self.ssbo_1, "4f 4x4 4f", ["in_vert", "in_col"])],
            mode=self.ctx.POINTS,
        )
        self.vao_2 = self.ctx.geometry(
            [BufferDescription(self.ssbo_2, "4f 4x4 4f", ["in_vert", "in_col"])],
            mode=self.ctx.POINTS,
        )

        # Program for visualizing the balls.
        # This is simply
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            // Input from buffers as described in the geometry
            in vec4 in_vert;
            in vec4 in_col;

            // Output values to the geometry shader
            out vec2 v_pos;
            out float v_radius;
            out vec4 v_col;

            void main()
            {
                // We just forward every every every ball to geometry shader
                v_pos = in_vert.xy;
                v_radius = in_vert.w;
                v_col = in_col;
            }
            """,
            geometry_shader="""
            #version 330

            // Configure the geometry shader to only take one vertex
            // per invocation from the vertex shader.
            layout (points) in;
            // Configure the geometry shader to emit triangle strips
            // and limit the max number of vertices to 4.
            layout (triangle_strip, max_vertices = 4) out;

            // Use pyglet's global projection UBO
            uniform WindowBlock {
                mat4 projection;
                mat4 view;
            } window;

            // The input from the vertex shader.
            // These should be unsized arrays since the geometry shader
            // can be configured to take more than one vertex from the
            // vertex shader per invocation. For example lines or triangles.
            // The compiler will ensure the arrays have the right size.
            in vec2 v_pos[];
            in vec4 v_col[];
            in float v_radius[];

            // Output color texture coordinates and color to the
            // fragment shader. These are emitted together with
            // the gl_Position every time EmitVertex() is called.
            out vec2 g_uv;
            out vec3 g_col;

            void main() {
                // We use the ball position as the center point of the rectangle we are emitting
                vec2 center = v_pos[0];
                // Calculate the half size to easily emit 4 vertices around the center.
                vec2 hsize = vec2(v_radius[0]);

                // Configure out color. This color can change per vertex
                // but we're using the same color for all vertices of the ball
                // so we only need to set it once.
                g_col = v_col[0].rgb;

                // Emit 4 vertices making a rectangle as a triangle strip.
                // We need to supply separate texture coordinates for each vertex
                gl_Position = window.projection * vec4(vec2(-hsize.x, hsize.y) + center, 0.0, 1.0);
                g_uv = vec2(0, 1);
                EmitVertex();
                gl_Position = window.projection * vec4(vec2(-hsize.x, -hsize.y) + center, 0.0, 1.0);
                g_uv = vec2(0, 0);
                EmitVertex();
                gl_Position = window.projection * vec4(vec2(hsize.x, hsize.y) + center, 0.0, 1.0);
                g_uv = vec2(1, 1);
                EmitVertex();
                gl_Position = window.projection * vec4(vec2(hsize.x, -hsize.y) + center, 0.0, 1.0);
                g_uv = vec2(1, 0);
                EmitVertex();

                // End the triangle strip here
                EndPrimitive();
            }

            """,
            fragment_shader="""
            #version 330

            // Inputs from the geometry shader.
            // The fragment shader runs for every pixel it needs to fill in a triangle.
            // These inputs will be an interpolated value based on the distance
            // from each vertex. See. barycentric coordinate.
            in vec2 g_uv;
            in vec3 g_col;

            // The pixel value that will end up on the screen
            out vec4 out_color;

            void main()
            {
                // We use the texture coordinates to calculate the distance
                // from the center of the rectangle to discard pixels
                // outside a certain radius so we end up with filled circle.
                if (length(vec2(0.5, 0.5) - g_uv.xy) > 0.25)
                {
                    // Stop this shader invocation.
                    // No fragment will be written to the screen.
                    discard;
                }
                // Set final pixel value. This is simply
                // the color from the geometry shader.
                out_color = vec4(g_col, 1.0);
            }
            """,
        )

    def on_draw(self):
        self.clear()
        self.ctx.disable(self.ctx.BLEND)

        # Calculate a directional force based on time.
        # This will affect the locations of the balls,
        force = math.sin(self.time / 10) / 2, math.cos(self.time / 10) / 2

        # Bind buffers to the binding indices configured in the compute shader
        self.ssbo_1.bind_to_storage_buffer(binding=0)
        self.ssbo_2.bind_to_storage_buffer(binding=1)

        # Set uniforms / global values in shader
        self.cs["screen_size"] = self.get_size()
        self.cs["force"] = force
        self.cs["frame_time"] = self.frame_time

        # Run the compute shader.
        # This will look at each ball in the first buffer
        # and write new balls in the second buffer applying
        # new positions, velocities and apply force.
        self.cs.run(group_x=self.work_groups_to_execute)

        # Draw the points and balls emitting circles to the screen.
        # We are drawing the balls in the second buffer.
        self.vao_2.render(self.program)

        # Swap the buffers around (we are ping-ping rendering between two buffers)
        self.ssbo_1, self.ssbo_2 = self.ssbo_2, self.ssbo_1
        # Swap what geometry we draw
        self.vao_1, self.vao_2 = self.vao_2, self.vao_1

    def on_update(self, delta_time: float):
        # Keep our time variables up to date
        self.frame_time = delta_time

    def gen_initial_data(self, num_balls):
        """
        Generate positions, velocities and color data
        for num_balls. This is a generator function
        that yields a stream of values.

        Each ball has:
        * position and radius (3 floats + 1 padding float)
        * velocity (2 floats + 2 padding floats)
        * color: (3 floats + 1 padding float)
        """
        for i in range(num_balls):
            angle = (i / self.num_balls) * math.pi * 2.0
            radius = random.random() * 10 + 3
            # pos
            yield random.randrange(0, self.width)
            yield random.randrange(0, self.height)
            yield 0.0  # z (padding)
            yield radius  # w / radius
            # Velocity
            v = random.random() * 1.0 + 0.1
            yield math.cos(angle) * v  # vx
            yield math.sin(angle) * v  # vy
            yield 0.0  # vz (padding)
            yield 0.0  # vw (padding)
            # Color
            yield 1.0 * random.random()  # r
            yield 1.0 * random.random()  # g
            yield 1.0 * random.random()  # b
            yield 1.0  # a


app = App()
arcade.run()
